from hazelcast.serialization.api import IdentifiedDataSerializable


_AGGREGATORS_FACTORY_ID = -29


class Aggregator(object):
    """Marker base class for all aggregators.

    Aggregators allow computing a value of some function (e.g sum or max) over
    the stored map entries. The computation is performed in a fully distributed
    manner, so no data other than the computed value is transferred to the client,
    making the computation fast.
    """

    pass


class _AbstractAggregator(Aggregator, IdentifiedDataSerializable):
    def __init__(self, attribute_path=None):
        self._attribute_path = attribute_path

    def write_data(self, object_data_output):
        raise NotImplementedError("write_data")

    def read_data(self, object_data_input):
        pass

    def get_factory_id(self):
        return _AGGREGATORS_FACTORY_ID

    def get_class_id(self):
        raise NotImplementedError("get_class_id")


class _CountAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 4


class _DistinctValuesAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_int(0)

    def get_class_id(self):
        return 5


class _DoubleAverageAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_double(0)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 6


class _DoubleSumAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_double(0)

    def get_class_id(self):
        return 7


class _FixedPointSumAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 8


class _FloatingPointSumAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_double(0)

    def get_class_id(self):
        return 9


class _IntegerAverageAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_long(0)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 10


class _IntegerSumAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 11


class _LongAverageAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_long(0)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 12


class _LongSumAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 13


class _MaxAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_object(None)

    def get_class_id(self):
        return 14


class _MinAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_object(None)

    def get_class_id(self):
        return 15


class _NumberAverageAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_double(0)
        object_data_output.write_long(0)

    def get_class_id(self):
        return 16


class _MaxByAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_object(None)
        object_data_output.write_object(None)

    def get_class_id(self):
        return 17


class _MinByAggregator(_AbstractAggregator):
    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)
        object_data_output.write_object(None)
        object_data_output.write_object(None)

    def get_class_id(self):
        return 18


def count(attribute_path=None):
    """Creates an aggregator that counts the input values.

    Accepts ``None`` input values and ``None`` extracted values.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[int]: An aggregator that counts the input values.
    """
    return _CountAggregator(attribute_path)


def distinct(attribute_path=None):
    """Creates an aggregator that calculates the distinct set of input values.

    Accepts ``None`` input values and ``None`` extracted values.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[set]: An aggregator that calculates the distinct set of
        input values.
    """
    return _DistinctValuesAggregator(attribute_path)


def double_avg(attribute_path=None):
    """Creates an aggregator that calculates the average of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must be of type ``double`` (primitive or boxed) in Java or of a type
    that can be converted to that. That means, one should be able to use this
    aggregator with ``float`` or ``int`` values sent from the Python client
    unless they are out of range for ``double`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[float]: An aggregator that calculates the average of
        the input values.
    """
    return _DoubleAverageAggregator(attribute_path)


def double_sum(attribute_path=None):
    """Creates an aggregator that calculates the sum of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must be of type ``double`` (primitive or boxed) in Java or of a type
    that can be converted to that. That means, one should be able to use this
    aggregator with ``float`` or ``int`` values sent from the Python client
    unless they are out of range for ``double`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[float]: An aggregator that calculates the sum of the input
        values.
    """
    return _DoubleSumAggregator(attribute_path)


def fixed_point_sum(attribute_path=None):
    """Creates an aggregator that calculates the sum of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Accepts generic number input values. That means, one should be able to
    use this aggregator with ``float`` or ``int`` value sent from the Python
    client unless they are out of range for ``long`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[int]: An aggregator that calculates the sum of the
        input values.
    """
    return _FixedPointSumAggregator(attribute_path)


def floating_point_sum(attribute_path=None):
    """Creates an aggregator that calculates the sum of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Accepts generic number input values. That means, one should be able to
    use this aggregator with ``float`` or ``int`` value sent from the Python
    client unless they are out of range for ``double`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[float]: An aggregator that calculates the sum of the
        input values.
    """
    return _FloatingPointSumAggregator(attribute_path)


def int_avg(attribute_path=None):
    """Creates an aggregator that calculates the average of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must be of type ``int`` (primitive or boxed) in Java or of a type
    that can be converted to that. That means, one should be able to use this
    aggregator with ``int`` values sent from the Python client unless they
    are out of range for ``int`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[int]: An aggregator that calculates the average of the
        input values.
    """
    return _IntegerAverageAggregator(attribute_path)


def int_sum(attribute_path=None):
    """Creates an aggregator that calculates the sum of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must be of type ``int`` (primitive or boxed) in Java or of a type
    that can be converted to that. That means, one should be able to use this
    aggregator with ``int`` values sent from the Python client unless they
    are out of range for ``int`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[int]: An aggregator that calculates the sum of the
        input values.
    """
    return _IntegerSumAggregator(attribute_path)


def long_avg(attribute_path=None):
    """Creates an aggregator that calculates the average of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must be of type ``long`` (primitive or boxed) in Java or of a type
    that can be converted to that. That means, one should be able to use this
    aggregator with ``int`` values sent from the Python client unless they
    are out of range for ``long`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[int]: An aggregator that calculates the average of the
        input values.
    """
    return _LongAverageAggregator(attribute_path)


def long_sum(attribute_path=None):
    """Creates an aggregator that calculates the sum of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must be of type ``long`` (primitive or boxed) in Java or of a type
    that can be converted to that. That means, one should be able to use this
    aggregator with ``int`` values sent from the Python client unless they
    are out of range for ``long`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[int]: An aggregator that calculates the sum of the
        input values.
    """
    return _LongSumAggregator(attribute_path)


def max_(attribute_path=None):
    """Creates an aggregator that calculates the max of the input values.

    Accepts ``None`` input values and ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must implement the ``Comparable`` interface in Java. That means, one
    should be able to use this aggregator with most of the primitive values
    sent from the Python client, as Java implements this interface for the
    equivalents of types like ``int``, ``str``, and ``float``.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[any]: An aggregator that calculates the max of the input
        values.
    """
    return _MaxAggregator(attribute_path)


def min_(attribute_path=None):
    """Creates an aggregator that calculates the min of the input values.

    Accepts ``None`` input values and ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must implement the ``Comparable`` interface in Java. That means, one
    should be able to use this aggregator with most of the primitive values
    sent from the Python client, as Java implements this interface for the
    equivalents of types like ``int``, ``str``, and ``float``.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[any]: An aggregator that calculates the min of the
        input values.
    """
    return _MinAggregator(attribute_path)


def number_avg(attribute_path=None):
    """Creates an aggregator that calculates the average of the input values.

    Does NOT accept ``None`` input values or ``None`` extracted values.

    Accepts generic number input values. That means, one should be able to
    use this aggregator with ``float`` or ``int`` value sent from the Python
    client unless they are out of range for ``double`` type in Java.

    Args:
        attribute_path (str): Extracts values from this path, if given.

    Returns:
        Aggregator[float]: An aggregator that calculates the average of
        the input values.
    """
    return _NumberAverageAggregator(attribute_path)


def max_by(attribute_path=None):
    """Creates an aggregator that calculates the max of the input values
    extracted from the given ``attribute_path`` and returns the input
    item containing the maximum value. If multiple items contain the
    maximum value, any of them is returned.

    Accepts ``None`` input values and ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must implement the ``Comparable`` interface in Java. That means, one
    should be able to use this aggregator with most of the primitive values
    sent from the Python client, as Java implements this interface for the
    equivalents of types like ``int``, ``str``, and ``float``.

    Args:
        attribute_path (str): Path to extract values from.

    Returns:
        Aggregator[hazelcast.core.MapEntry]: An aggregator that calculates
        the input value containing the maximum value extracted from the path.
    """
    return _MaxByAggregator(attribute_path)


def min_by(attribute_path):
    """Creates an aggregator that calculates the min of the input values
    extracted from the given ``attribute_path`` and returns the input
    item containing the minimum value. If multiple items contain the
    minimum value, any of them is returned.

    Accepts ``None`` input values and ``None`` extracted values.

    Since the server-side implementation is in Java, values stored in the
    Map must implement the ``Comparable`` interface in Java. That means, one
    should be able to use this aggregator with most of the primitive values
    sent from the Python client, as Java implements this interface for the
    equivalents of types like ``int``, ``str``, and ``float``.

    Args:
        attribute_path (str): Path to extract values from.

    Returns:
        Aggregator[hazelcast.core.MapEntry]: An aggregator that calculates
        the input value containing the minimum value extracted from the path.
    """
    return _MinByAggregator(attribute_path)



from hazelcast.errors import HazelcastSerializationError
from hazelcast import six


class FieldType(object):
    PORTABLE = 0
    BYTE = 1
    BOOLEAN = 2
    CHAR = 3
    SHORT = 4
    INT = 5
    LONG = 6
    FLOAT = 7
    DOUBLE = 8
    UTF = 9  # Defined for backward compatibility.
    STRING = 9
    PORTABLE_ARRAY = 10
    BYTE_ARRAY = 11
    BOOLEAN_ARRAY = 12
    CHAR_ARRAY = 13
    SHORT_ARRAY = 14
    INT_ARRAY = 15
    LONG_ARRAY = 16
    FLOAT_ARRAY = 17
    DOUBLE_ARRAY = 18
    UTF_ARRAY = 19  # Defined for backward compatibility.
    STRING_ARRAY = 19


class FieldDefinition(object):
    def __init__(self, index, field_name, field_type, version, factory_id=0, class_id=0):
        self.index = index
        self.field_name = field_name
        self.field_type = field_type
        self.version = version
        self.factory_id = factory_id
        self.class_id = class_id

    def __eq__(self, other):
        return (
            isinstance(other, FieldDefinition)
            and self.index == other.index
            and self.field_name == other.field_name
            and self.version == other.version
            and self.factory_id == other.factory_id
            and self.class_id == other.class_id
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "FieldDefinition(ix=%s, name=%s, type=%s, version=%s, fid=%s, cid=%s)" % (
            self.index,
            self.field_name,
            self.field_type,
            self.version,
            self.factory_id,
            self.class_id,
        )


class ClassDefinition(object):
    def __init__(self, factory_id, class_id, version):
        self.factory_id = factory_id
        self.class_id = class_id
        self.version = version
        self.field_defs = {}  # string:FieldDefinition

    def add_field_def(self, field_def):
        self.field_defs[field_def.field_name] = field_def

    def get_field(self, field_name_or_index):
        if isinstance(field_name_or_index, int):
            index = field_name_or_index
            count = self.get_field_count()
            if 0 <= index < count:
                for field in six.itervalues(self.field_defs):
                    if field.index == index:
                        return field
            raise IndexError("Index is out of bound. Index: %s and size: %s" % (index, count))
        else:
            return self.field_defs.get(field_name_or_index, None)

    def has_field(self, field_name):
        return field_name in self.field_defs

    def get_field_names(self):
        return list(self.field_defs.keys())

    def get_field_type(self, field_name):
        fd = self.get_field(field_name)
        if fd:
            return fd.field_type
        raise ValueError("Unknown field: %s" % field_name)

    def get_field_class_id(self, field_name):
        fd = self.get_field(field_name)
        if fd:
            return fd.class_id
        raise ValueError("Unknown field: %s" % field_name)

    def get_field_count(self):
        return len(self.field_defs)

    def set_version_if_not_set(self, version):
        if self.version < 0:
            self.version = version

    def __eq__(self, other):
        return (
            isinstance(other, ClassDefinition)
            and self.factory_id == other.factory_id
            and self.class_id == other.class_id
            and self.version == other.version
            and self.field_defs == other.field_defs
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "fid:%s, cid:%s, v:%s, fields:%s" % (
            self.factory_id,
            self.class_id,
            self.version,
            self.field_defs,
        )

    def __hash__(self):
        return hash((self.factory_id, self.class_id, self.version))


class ClassDefinitionBuilder(object):
    """Builder class to construct :class:`ClassDefinition` of
    :class:`hazelcast.serialization.api.Portable` objects.

    One must make sure that the order of the fields added to this
    builder is consistent across all the usages of the Portable
    object such as the write order of the fields of the
    :func:`Portable.write_portable <hazelcast.serialization.api.Portable.write_portable>`
    method.
    """

    def __init__(self, factory_id, class_id, version=0):
        self.factory_id = factory_id
        self.class_id = class_id
        self.version = version

        self._index = 0
        self._done = False
        self._field_defs = list()
        self._field_names = set()

    def add_portable_field(self, field_name, class_def):
        """Adds the field with the Portable type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.
            class_def (ClassDefinition): Class definition
                of the nested Portable.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        if class_def.class_id is None or class_def.class_id == 0:
            raise ValueError("Portable class id cannot be zero!")
        self._add_field_by_type(
            field_name,
            FieldType.PORTABLE,
            class_def.version,
            class_def.factory_id,
            class_def.class_id,
        )
        return self

    def add_byte_field(self, field_name):
        """Adds the field with the byte type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.BYTE, self.version)
        return self

    def add_boolean_field(self, field_name):
        """Adds the field with the boolean type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.BOOLEAN, self.version)
        return self

    def add_char_field(self, field_name):
        """Adds the field with the char type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.CHAR, self.version)
        return self

    def add_short_field(self, field_name):
        """Adds the field with the short type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.SHORT, self.version)
        return self

    def add_int_field(self, field_name):
        """Adds the field with the int type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.INT, self.version)
        return self

    def add_long_field(self, field_name):
        """Adds the field with the long type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.LONG, self.version)
        return self

    def add_float_field(self, field_name):
        """Adds the field with the float type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.FLOAT, self.version)
        return self

    def add_double_field(self, field_name):
        """Adds the field with the double type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.DOUBLE, self.version)
        return self

    def add_string_field(self, field_name):
        """Adds the field with the string type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.STRING, self.version)
        return self

    def add_utf_field(self, field_name):
        """Adds the field with the string type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`add_string_field` instead.
        """
        return self.add_string_field(field_name)

    def add_portable_array_field(self, field_name, class_def):
        """Adds the field with the Portable array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.
            class_def (ClassDefinition): Class definition
                of the nested Portable.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        if class_def.class_id is None or class_def.class_id == 0:
            raise ValueError("Portable class id cannot be zero!")
        self._add_field_by_type(
            field_name,
            FieldType.PORTABLE_ARRAY,
            class_def.version,
            class_def.factory_id,
            class_def.class_id,
        )
        return self

    def add_byte_array_field(self, field_name):
        """Adds the field with the byte array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.BYTE_ARRAY, self.version)
        return self

    def add_boolean_array_field(self, field_name):
        """Adds the field with the boolean array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.BOOLEAN_ARRAY, self.version)
        return self

    def add_char_array_field(self, field_name):
        """Adds the field with the char array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.CHAR_ARRAY, self.version)
        return self

    def add_short_array_field(self, field_name):
        """Adds the field with the short array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.SHORT_ARRAY, self.version)
        return self

    def add_int_array_field(self, field_name):
        """Adds the field with the int array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.INT_ARRAY, self.version)
        return self

    def add_long_array_field(self, field_name):
        """Adds the field with the long array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.LONG_ARRAY, self.version)
        return self

    def add_float_array_field(self, field_name):
        """Adds the field with the float array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.FLOAT_ARRAY, self.version)
        return self

    def add_double_array_field(self, field_name):
        """Adds the field with the double array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.DOUBLE_ARRAY, self.version)
        return self

    def add_string_array_field(self, field_name):
        """Adds the field with the string array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.
        """
        self._add_field_by_type(field_name, FieldType.STRING_ARRAY, self.version)
        return self

    def add_utf_array_field(self, field_name):
        """Adds the field with the string array type to the
        class definition.

        Args:
            field_name (str): Name of the field to add.

        Returns:
            ClassDefinitionBuilder: Itself for chaining.

        Raises:
            HazelcastSerializationError: If this method is called
                after :func:`build` or a field with the same
                name is already registered.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`add_string_array_field` instead.
        """
        return self.add_string_array_field(field_name)

    def add_field_def(self, field_def):
        """
        Warnings:
              This method is not intended for public usage.
              It might be removed from the public API on the
              next major version.
        """
        if self._index != field_def.index:
            raise ValueError("Invalid field index")
        self._check(field_def.field_name)
        self._index += 1
        self._field_defs.append(field_def)
        return self

    def build(self):
        """Builds and returns the class definition.

        Returns:
            ClassDefinition:
        """
        self._done = True
        cd = ClassDefinition(self.factory_id, self.class_id, self.version)
        for field_def in self._field_defs:
            cd.add_field_def(field_def)
        return cd

    def _add_field_by_type(self, field_name, field_type, version, factory_id=0, class_id=0):
        self._check(field_name)
        fd = FieldDefinition(self._index, field_name, field_type, version, factory_id, class_id)
        self._field_defs.append(fd)
        self._index += 1

    def _check(self, field_name):
        if field_name in self._field_names:
            raise HazelcastSerializationError("Field with the name %s already exists" % field_name)
        self._field_names.add(field_name)

        if self._done:
            raise HazelcastSerializationError(
                "ClassDefinition is already built for %s" % self.class_id
            )



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.core import MemberVersion

_MAJOR_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_MAJOR_DECODE_OFFSET = 0
_MINOR_ENCODE_OFFSET = _MAJOR_ENCODE_OFFSET + BYTE_SIZE_IN_BYTES
_MINOR_DECODE_OFFSET = _MAJOR_DECODE_OFFSET + BYTE_SIZE_IN_BYTES
_PATCH_ENCODE_OFFSET = _MINOR_ENCODE_OFFSET + BYTE_SIZE_IN_BYTES
_PATCH_DECODE_OFFSET = _MINOR_DECODE_OFFSET + BYTE_SIZE_IN_BYTES
_INITIAL_FRAME_SIZE = _PATCH_ENCODE_OFFSET + BYTE_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class MemberVersionCodec(object):
    @staticmethod
    def encode(buf, member_version, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_byte(initial_frame_buf, _MAJOR_ENCODE_OFFSET, member_version.major)
        FixSizedTypesCodec.encode_byte(initial_frame_buf, _MINOR_ENCODE_OFFSET, member_version.minor)
        FixSizedTypesCodec.encode_byte(initial_frame_buf, _PATCH_ENCODE_OFFSET, member_version.patch)
        buf.extend(initial_frame_buf)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        major = FixSizedTypesCodec.decode_byte(initial_frame.buf, _MAJOR_DECODE_OFFSET)
        minor = FixSizedTypesCodec.decode_byte(initial_frame.buf, _MINOR_DECODE_OFFSET)
        patch = FixSizedTypesCodec.decode_byte(initial_frame.buf, _PATCH_DECODE_OFFSET)
        CodecUtil.fast_forward_to_end_frame(msg)
        return MemberVersion(major, minor, patch)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE, EVENT_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import CodecUtil

# hex: 0x0D0B00
_REQUEST_MESSAGE_TYPE = 854784
# hex: 0x0D0B01
_RESPONSE_MESSAGE_TYPE = 854785
# hex: 0x0D0B02
_EVENT_ENTRY_MESSAGE_TYPE = 854786

_REQUEST_LOCAL_ONLY_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_LOCAL_ONLY_OFFSET + BOOLEAN_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE
_EVENT_ENTRY_EVENT_TYPE_OFFSET = EVENT_HEADER_SIZE
_EVENT_ENTRY_UUID_OFFSET = _EVENT_ENTRY_EVENT_TYPE_OFFSET + INT_SIZE_IN_BYTES
_EVENT_ENTRY_NUMBER_OF_AFFECTED_ENTRIES_OFFSET = _EVENT_ENTRY_UUID_OFFSET + UUID_SIZE_IN_BYTES


def encode_request(name, predicate, local_only):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_LOCAL_ONLY_OFFSET, local_only)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, predicate, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)


def handle(msg, handle_entry_event=None):
    message_type = msg.get_message_type()
    if message_type == _EVENT_ENTRY_MESSAGE_TYPE and handle_entry_event is not None:
        initial_frame = msg.next_frame()
        event_type = FixSizedTypesCodec.decode_int(initial_frame.buf, _EVENT_ENTRY_EVENT_TYPE_OFFSET)
        uuid = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _EVENT_ENTRY_UUID_OFFSET)
        number_of_affected_entries = FixSizedTypesCodec.decode_int(initial_frame.buf, _EVENT_ENTRY_NUMBER_OF_AFFECTED_ENTRIES_OFFSET)
        key = CodecUtil.decode_nullable(msg, DataCodec.decode)
        value = CodecUtil.decode_nullable(msg, DataCodec.decode)
        old_value = CodecUtil.decode_nullable(msg, DataCodec.decode)
        merging_value = CodecUtil.decode_nullable(msg, DataCodec.decode)
        handle_entry_event(key, value, old_value, merging_value, event_type, uuid, number_of_affected_entries)
        return



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.core import SimpleEntryView
from hazelcast.protocol.builtin import DataCodec

_COST_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_COST_DECODE_OFFSET = 0
_CREATION_TIME_ENCODE_OFFSET = _COST_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_CREATION_TIME_DECODE_OFFSET = _COST_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_EXPIRATION_TIME_ENCODE_OFFSET = _CREATION_TIME_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_EXPIRATION_TIME_DECODE_OFFSET = _CREATION_TIME_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_HITS_ENCODE_OFFSET = _EXPIRATION_TIME_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_HITS_DECODE_OFFSET = _EXPIRATION_TIME_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_LAST_ACCESS_TIME_ENCODE_OFFSET = _HITS_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_LAST_ACCESS_TIME_DECODE_OFFSET = _HITS_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_LAST_STORED_TIME_ENCODE_OFFSET = _LAST_ACCESS_TIME_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_LAST_STORED_TIME_DECODE_OFFSET = _LAST_ACCESS_TIME_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_LAST_UPDATE_TIME_ENCODE_OFFSET = _LAST_STORED_TIME_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_LAST_UPDATE_TIME_DECODE_OFFSET = _LAST_STORED_TIME_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_VERSION_ENCODE_OFFSET = _LAST_UPDATE_TIME_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_VERSION_DECODE_OFFSET = _LAST_UPDATE_TIME_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_TTL_ENCODE_OFFSET = _VERSION_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_TTL_DECODE_OFFSET = _VERSION_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_MAX_IDLE_ENCODE_OFFSET = _TTL_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_MAX_IDLE_DECODE_OFFSET = _TTL_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_INITIAL_FRAME_SIZE = _MAX_IDLE_ENCODE_OFFSET + LONG_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class SimpleEntryViewCodec(object):
    @staticmethod
    def encode(buf, simple_entry_view, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _COST_ENCODE_OFFSET, simple_entry_view.cost)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _CREATION_TIME_ENCODE_OFFSET, simple_entry_view.creation_time)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _EXPIRATION_TIME_ENCODE_OFFSET, simple_entry_view.expiration_time)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _HITS_ENCODE_OFFSET, simple_entry_view.hits)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _LAST_ACCESS_TIME_ENCODE_OFFSET, simple_entry_view.last_access_time)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _LAST_STORED_TIME_ENCODE_OFFSET, simple_entry_view.last_stored_time)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _LAST_UPDATE_TIME_ENCODE_OFFSET, simple_entry_view.last_update_time)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _VERSION_ENCODE_OFFSET, simple_entry_view.version)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _TTL_ENCODE_OFFSET, simple_entry_view.ttl)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _MAX_IDLE_ENCODE_OFFSET, simple_entry_view.max_idle)
        buf.extend(initial_frame_buf)
        DataCodec.encode(buf, simple_entry_view.key)
        DataCodec.encode(buf, simple_entry_view.value)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        cost = FixSizedTypesCodec.decode_long(initial_frame.buf, _COST_DECODE_OFFSET)
        creation_time = FixSizedTypesCodec.decode_long(initial_frame.buf, _CREATION_TIME_DECODE_OFFSET)
        expiration_time = FixSizedTypesCodec.decode_long(initial_frame.buf, _EXPIRATION_TIME_DECODE_OFFSET)
        hits = FixSizedTypesCodec.decode_long(initial_frame.buf, _HITS_DECODE_OFFSET)
        last_access_time = FixSizedTypesCodec.decode_long(initial_frame.buf, _LAST_ACCESS_TIME_DECODE_OFFSET)
        last_stored_time = FixSizedTypesCodec.decode_long(initial_frame.buf, _LAST_STORED_TIME_DECODE_OFFSET)
        last_update_time = FixSizedTypesCodec.decode_long(initial_frame.buf, _LAST_UPDATE_TIME_DECODE_OFFSET)
        version = FixSizedTypesCodec.decode_long(initial_frame.buf, _VERSION_DECODE_OFFSET)
        ttl = FixSizedTypesCodec.decode_long(initial_frame.buf, _TTL_DECODE_OFFSET)
        max_idle = FixSizedTypesCodec.decode_long(initial_frame.buf, _MAX_IDLE_DECODE_OFFSET)
        key = DataCodec.decode(msg)
        value = DataCodec.decode(msg)
        CodecUtil.fast_forward_to_end_frame(msg)
        return SimpleEntryView(key, value, cost, creation_time, expiration_time, hits, last_access_time, last_stored_time, last_update_time, version, ttl, max_idle)



from hazelcast.protocol.codec import (
    transactional_map_contains_key_codec,
    transactional_map_delete_codec,
    transactional_map_get_codec,
    transactional_map_get_for_update_codec,
    transactional_map_is_empty_codec,
    transactional_map_key_set_codec,
    transactional_map_key_set_with_predicate_codec,
    transactional_map_put_codec,
    transactional_map_put_if_absent_codec,
    transactional_map_remove_codec,
    transactional_map_remove_if_same_codec,
    transactional_map_replace_codec,
    transactional_map_replace_if_same_codec,
    transactional_map_set_codec,
    transactional_map_size_codec,
    transactional_map_values_codec,
    transactional_map_values_with_predicate_codec,
)
from hazelcast.proxy.base import TransactionalProxy
from hazelcast.util import check_not_none, to_millis, thread_id, ImmutableLazyDataList


class TransactionalMap(TransactionalProxy):
    """Transactional implementation of :class:`~hazelcast.proxy.map.Map`."""

    def contains_key(self, key):
        """Transactional implementation of :func:`Map.contains_key(key) <hazelcast.proxy.map.Map.contains_key>`

        Args:
            key: The specified key.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this map contains an entry for the specified key,
            ``False`` otherwise.
        """
        check_not_none(key, "key can't be none")
        key_data = self._to_data(key)
        request = transactional_map_contains_key_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request, transactional_map_contains_key_codec.decode_response)

    def get(self, key):
        """Transactional implementation of :func:`Map.get(key) <hazelcast.proxy.map.Map.get>`

        Args:
            key: The specified key.

        Returns:
            hazelcast.future.Future[any]: The value for the specified key.
        """
        check_not_none(key, "key can't be none")

        def handler(message):
            return self._to_object(transactional_map_get_codec.decode_response(message))

        key_data = self._to_data(key)
        request = transactional_map_get_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request, handler)

    def get_for_update(self, key):
        """Locks the key and then gets and returns the value to which the specified key is mapped.

        Lock will be released at the end of the transaction (either commit or rollback).

        Args:
            key: The specified key.

        Returns:
            hazelcast.future.Future[any]: The value for the specified key.

        See Also:
            :func:`Map.get(key) <hazelcast.proxy.map.Map.get>`
        """
        check_not_none(key, "key can't be none")

        def handler(message):
            return self._to_object(transactional_map_get_for_update_codec.decode_response(message))

        key_data = self._to_data(key)
        request = transactional_map_get_for_update_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request, handler)

    def size(self):
        """Transactional implementation of :func:`Map.size() <hazelcast.proxy.map.Map.size>`

        Returns:
            hazelcast.future.Future[int]: Number of entries in this map.
        """
        request = transactional_map_size_codec.encode_request(
            self.name, self.transaction.id, thread_id()
        )
        return self._invoke(request, transactional_map_size_codec.decode_response)

    def is_empty(self):
        """Transactional implementation of :func:`Map.is_empty() <hazelcast.proxy.map.Map.is_empty>`

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this map contains no key-value mappings, ``False`` otherwise.
        """
        request = transactional_map_is_empty_codec.encode_request(
            self.name, self.transaction.id, thread_id()
        )
        return self._invoke(request, transactional_map_is_empty_codec.decode_response)

    def put(self, key, value, ttl=None):
        """Transactional implementation of :func:`Map.put(key, value, ttl) <hazelcast.proxy.map.Map.put>`

        The object to be put will be accessible only in the current transaction context till the transaction is
        committed.

        Args:
            key: The specified key.
            value: The value to associate with the key.
            ttl (int): Maximum time in seconds for this entry to stay.

        Returns:
            hazelcast.future.Future[any]: Previous value associated with key or ``None``
            if there was no mapping for key.
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")

        def handler(message):
            return self._to_object(transactional_map_put_codec.decode_response(message))

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = transactional_map_put_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, value_data, to_millis(ttl)
        )
        return self._invoke(request, handler)

    def put_if_absent(self, key, value):
        """Transactional implementation of :func:`Map.put_if_absent(key, value) <hazelcast.proxy.map.Map.put_if_absent>`

        The object to be put will be accessible only in the current transaction context till the transaction is
        committed.

        Args:
            key: Key of the entry.
            value: Value of the entry.

        Returns:
            hazelcast.future.Future[any]: Old value of the entry.
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")

        def handler(message):
            return self._to_object(transactional_map_put_if_absent_codec.decode_response(message))

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = transactional_map_put_if_absent_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, value_data
        )
        return self._invoke(request, handler)

    def set(self, key, value):
        """Transactional implementation of :func:`Map.set(key, value) <hazelcast.proxy.map.Map.set>`

        The object to be set will be accessible only in the current transaction context till the transaction is
        committed.

        Args:
            key: Key of the entry.
            value: Value of the entry.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = transactional_map_set_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, value_data
        )
        return self._invoke(request)

    def replace(self, key, value):
        """Transactional implementation of :func:`Map.replace(key, value) <hazelcast.proxy.map.Map.replace>`

        The object to be replaced will be accessible only in the current transaction context till the transaction is
        committed.

        Args:
            key: The specified key.
            value: The value to replace the previous value.

        Returns:
            hazelcast.future.Future[any]: Previous value associated with key, or ``None``
            if there was no mapping for key.
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")

        def handler(message):
            return self._to_object(transactional_map_replace_codec.decode_response(message))

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = transactional_map_replace_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, value_data
        )
        return self._invoke(request, handler)

    def replace_if_same(self, key, old_value, new_value):
        """Transactional implementation of :func:`Map.replace_if_same(key, old_value, new_value)
        <hazelcast.proxy.map.Map.replace_if_same>`

        The object to be replaced will be accessible only in the current transaction context till the transaction is
        committed.

        Args:
            key: The specified key.
            old_value: Replace the key value if it is the old value.
            new_value: The new value to replace the old value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the value was replaced, ``False`` otherwise.
        """
        check_not_none(key, "key can't be none")
        check_not_none(old_value, "old_value can't be none")
        check_not_none(new_value, "new_value can't be none")

        key_data = self._to_data(key)
        old_value_data = self._to_data(old_value)
        new_value_data = self._to_data(new_value)
        request = transactional_map_replace_if_same_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, old_value_data, new_value_data
        )
        return self._invoke(request, transactional_map_replace_if_same_codec.decode_response)

    def remove(self, key):
        """Transactional implementation of :func:`Map.remove(key) <hazelcast.proxy.map.Map.remove>`

        The object to be removed will be removed from only the current transaction context until the transaction is
        committed.

        Args:
            key: Key of the mapping to be deleted.

        Returns:
            hazelcast.future.Future[any]: The previous value associated with key, or ``None``
            if there was no mapping for key.
        """
        check_not_none(key, "key can't be none")

        def handler(message):
            return self._to_object(transactional_map_remove_codec.decode_response(message))

        key_data = self._to_data(key)
        request = transactional_map_remove_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request, handler)

    def remove_if_same(self, key, value):
        """Transactional implementation of :func:`Map.remove_if_same(key, value)
        <hazelcast.proxy.map.Map.remove_if_same>`

        The object to be removed will be removed from only the current transaction context until the transaction is
        committed.

        Args:
            key: The specified key.
            value: Remove the key if it has this value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the value was removed, ``False`` otherwise.
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = transactional_map_remove_if_same_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, value_data
        )
        return self._invoke(request, transactional_map_remove_if_same_codec.decode_response)

    def delete(self, key):
        """Transactional implementation of :func:`Map.delete(key) <hazelcast.proxy.map.Map.delete>`

        The object to be deleted will be removed from only the current transaction context until the transaction is
        committed.

        Args:
            key: Key of the mapping to be deleted.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be none")

        key_data = self._to_data(key)
        request = transactional_map_delete_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request)

    def key_set(self, predicate=None):
        """Transactional implementation of :func:`Map.key_set(predicate) <hazelcast.proxy.map.Map.key_set>`

        Args:
            predicate (hazelcast.predicate.Predicate): Predicate to filter the entries.

        Returns:
            hazelcast.future.Future[list]: A list of the clone of the keys.
        """
        if predicate:

            def handler(message):
                return ImmutableLazyDataList(
                    transactional_map_key_set_with_predicate_codec.decode_response(message),
                    self._to_object,
                )

            predicate_data = self._to_data(predicate)
            request = transactional_map_key_set_with_predicate_codec.encode_request(
                self.name, self.transaction.id, thread_id(), predicate_data
            )
        else:

            def handler(message):
                return ImmutableLazyDataList(
                    transactional_map_key_set_codec.decode_response(message), self._to_object
                )

            request = transactional_map_key_set_codec.encode_request(
                self.name, self.transaction.id, thread_id()
            )

        return self._invoke(request, handler)

    def values(self, predicate=None):
        """Transactional implementation of :func:`Map.values(predicate) <hazelcast.proxy.map.Map.values>`

        Args:
            predicate (hazelcast.predicate.Predicate): Predicate to filter the entries.

        Returns:
            hazelcast.future.Future[list]: A list of clone of the values contained in this map.
        """
        if predicate:

            def handler(message):
                return ImmutableLazyDataList(
                    transactional_map_values_with_predicate_codec.decode_response(message),
                    self._to_object,
                )

            predicate_data = self._to_data(predicate)
            request = transactional_map_values_with_predicate_codec.encode_request(
                self.name, self.transaction.id, thread_id(), predicate_data
            )
        else:

            def handler(message):
                return ImmutableLazyDataList(
                    transactional_map_values_codec.decode_response(message), self._to_object
                )

            request = transactional_map_values_codec.encode_request(
                self.name, self.transaction.id, thread_id()
            )

        return self._invoke(request, handler)



import itertools
import threading
import collections

from hazelcast.proxy.base import Proxy, MAX_SIZE
from hazelcast.config import _FlakeIdGeneratorConfig
from hazelcast.util import current_time
from hazelcast.protocol.codec import flake_id_generator_new_id_batch_codec
from hazelcast.future import ImmediateFuture, Future


class FlakeIdGenerator(Proxy):
    """A cluster-wide unique ID generator. Generated IDs are int (long in case of the Python 2 on 32 bit architectures)
    values and are k-ordered (roughly ordered). IDs are in the range from 0 to 2^63 - 1.

    The IDs contain timestamp component and a node ID component, which is assigned when the member
    joins the cluster. This allows the IDs to be ordered and unique without any coordination between
    members, which makes the generator safe even in split-brain scenario.

    Timestamp component is in milliseconds since 1.1.2018, 0:00 UTC and has 41 bits. This caps
    the useful lifespan of the generator to little less than 70 years (until ~2088). The sequence component
    is 6 bits. If more than 64 IDs are requested in single millisecond, IDs will gracefully overflow to the next
    millisecond and uniqueness is guaranteed in this case. The implementation does not allow overflowing
    by more than 15 seconds, if IDs are requested at higher rate, the call will block. Note, however, that
    clients are able to generate even faster because each call goes to a different (random) member and
    the 64 IDs/ms limit is for single member.

    Node ID overflow:
        It is possible to generate IDs on any member or client as long as there is at least one
        member with join version smaller than 2^16 in the cluster. The remedy is to restart the cluster:
        nodeId will be assigned from zero again. Uniqueness after the restart will be preserved thanks to
        the timestamp component.
    """

    _BITS_NODE_ID = 16
    _BITS_SEQUENCE = 6

    def __init__(self, service_name, name, context):
        super(FlakeIdGenerator, self).__init__(service_name, name, context)

        config = context.config.flake_id_generators.get(name, None)
        if config is None:
            config = _FlakeIdGeneratorConfig()

        self._auto_batcher = _AutoBatcher(
            config.prefetch_count, config.prefetch_validity, self._new_id_batch
        )

    def new_id(self):
        """Generates and returns a cluster-wide unique ID.

        This method goes to a random member and gets a batch of IDs, which will then be returned
        locally for limited time. The pre-fetch size and the validity time can be configured.

        Note:
            Values returned from this method may not be strictly ordered.

        Returns:
          hazelcast.future.Future[int], new cluster-wide unique ID.

        Raises:
            HazelcastError: if node ID for all members in the cluster is out of valid range.
                See ``Node ID overflow`` note above.
        """
        return self._auto_batcher.new_id()

    def _new_id_batch(self, batch_size):
        def handler(message):
            response = flake_id_generator_new_id_batch_codec.decode_response(message)
            return _IdBatch(response["base"], response["increment"], response["batch_size"])

        request = flake_id_generator_new_id_batch_codec.encode_request(self.name, batch_size)
        return self._invoke(request, handler)


class _AutoBatcher(object):
    def __init__(self, batch_size, validity, id_generator):
        self._batch_size = batch_size
        self._validity = validity
        self._batch_id_supplier = id_generator
        self._block = _Block(_IdBatch(0, 0, 0), 0)
        self._lock = threading.RLock()
        self._id_queue = collections.deque()
        self._request_in_air = False

    def new_id(self):
        while True:
            block = self._block
            next_id = block.next_id()
            if next_id is not None:
                return ImmediateFuture(next_id)

            with self._lock:
                # new block was assigned in the meantime
                if block is not self._block:
                    continue

                future = Future()
                self._id_queue.append(future)
                if not self._request_in_air:
                    self._request_in_air = True
                    self._request_new_batch()
                return future

    def _request_new_batch(self):
        future = self._batch_id_supplier(self._batch_size)
        future.add_done_callback(self._assign_new_block)

    def _assign_new_block(self, future):
        try:
            new_batch_required = False
            id_batch = future.result()
            block = _Block(id_batch, self._validity)
            with self._lock:
                while True:
                    try:
                        f = self._id_queue.popleft()
                        next_id = block.next_id()
                        if next_id is not None:
                            f.set_result(next_id)
                        else:
                            self._id_queue.appendleft(f)
                            new_batch_required = True
                            break
                    except IndexError:
                        break
                if new_batch_required:
                    self._request_in_air = True
                    self._request_new_batch()
                else:
                    self._request_in_air = False
                    self._block = block
        except Exception as ex:
            with self._lock:
                while True:
                    try:
                        f = self._id_queue.popleft()
                        f.set_exception(ex)
                    except IndexError:
                        break
                self._request_in_air = False


class _IdBatch(object):
    def __init__(self, base, increment, batch_size):
        self._base = base
        self._increment = increment
        self._batch_size = batch_size

    def __iter__(self):
        self._remaining = itertools.count(self._batch_size, -1)
        self._next_id = itertools.count(self._base, self._increment)
        return self

    def __next__(self):
        if next(self._remaining) <= 0:
            raise StopIteration

        return next(self._next_id)

    # For Python 2 compatibility
    next = __next__


class _Block(object):
    def __init__(self, id_batch, validity):
        self._id_batch = id_batch
        self._iterator = iter(self._id_batch)
        self._invalid_since = validity + current_time() if validity > 0 else MAX_SIZE

    def next_id(self):
        if self._invalid_since <= current_time():
            return None

        return next(self._iterator, None)



import hazelcast.util as util
from hazelcast.errors import HazelcastSerializationError
from hazelcast.serialization.api import StreamSerializer, Portable
from hazelcast.serialization.portable.reader import DefaultPortableReader, MorphingPortableReader
from hazelcast.serialization.portable.writer import DefaultPortableWriter
from hazelcast.serialization.serialization_const import CONSTANT_TYPE_PORTABLE


class PortableSerializer(StreamSerializer):
    def __init__(self, portable_context, portable_factories):
        self._portable_context = portable_context
        self._portable_factories = portable_factories

    def write(self, out, portable):
        if not isinstance(portable, Portable):
            raise ValueError("Input object must be an instance of Portable")
        if portable.get_class_id() == 0:
            raise ValueError("Portable class id cannot be zero!")
        out.write_int(portable.get_factory_id())
        out.write_int(portable.get_class_id())
        self.write_internal(out, portable)

    def write_internal(self, out, portable):
        cd = self._portable_context.lookup_or_register_class_definition(portable)
        out.write_int(cd.version)

        writer = DefaultPortableWriter(self, out, cd)
        portable.write_portable(writer)
        writer.end()

    def read(self, inp):
        factory_id = inp.read_int()
        class_id = inp.read_int()
        return self.read_internal(inp, factory_id, class_id)

    def read_internal(self, inp, factory_id, class_id):
        version = inp.read_int()
        portable = self.create_new_portable_instance(factory_id, class_id)
        portable_version = self.find_portable_version(factory_id, class_id, portable)
        reader = self.create_reader(inp, factory_id, class_id, version, portable_version)
        portable.read_portable(reader)
        reader.end()
        return portable

    def find_portable_version(self, factory_id, class_id, portable):
        current_version = self._portable_context.get_class_version(factory_id, class_id)
        if current_version < 0:
            current_version = util.get_portable_version(
                portable, self._portable_context.portable_version
            )
        if current_version > 0:
            self._portable_context.set_class_version(factory_id, class_id, current_version)
        return current_version

    def create_new_portable_instance(self, factory_id, class_id):
        try:
            portable_factory = self._portable_factories[factory_id]
        except KeyError:
            raise HazelcastSerializationError(
                "Could not find portable_factory for factory-id: %s" % factory_id
            )

        portable = portable_factory[class_id]
        if portable is None:
            raise HazelcastSerializationError(
                "Could not create Portable for class-id: %s" % class_id
            )
        return portable()

    def create_reader(self, inp, factory_id, class_id, version, portable_version):
        effective_version = version
        if version < 0:
            effective_version = self._portable_context.getVersion()

        cd = self._portable_context.lookup_class_definition(factory_id, class_id, effective_version)
        if cd is None:
            begin = inp.position()
            cd = self._portable_context.read_class_definition(
                inp, factory_id, class_id, effective_version
            )
            inp.set_position(begin)

        if portable_version == effective_version:
            reader = DefaultPortableReader(self, inp, cd)
        else:
            reader = MorphingPortableReader(self, inp, cd)
        return reader

    def create_morphing_reader(self, inp):
        factory_id = inp.read_int()
        class_id = inp.read_int()
        version = inp.read_int()

        portable = self.create_new_portable_instance(factory_id, class_id)
        portable_version = self.find_portable_version(factory_id, class_id, portable)

        return self.create_reader(inp, factory_id, class_id, version, portable_version)

    def create_default_reader(self, inp):
        factory_id = inp.read_int()
        class_id = inp.read_int()
        version = inp.read_int()

        return self.create_reader(inp, factory_id, class_id, version, version)

    def get_type_id(self):
        return CONSTANT_TYPE_PORTABLE

    def destroy(self):
        self._portable_factories.clear()



import threading

from hazelcast import util
from hazelcast.errors import HazelcastSerializationError
from hazelcast.serialization import bits
from hazelcast.serialization.portable.classdef import (
    ClassDefinition,
    ClassDefinitionBuilder,
    FieldType,
    FieldDefinition,
)
from hazelcast.serialization.portable.writer import ClassDefinitionWriter
from hazelcast.six.moves import range


class PortableContext(object):
    def __init__(self, serialization_service, portable_version):
        self.serialization_service = serialization_service
        self.portable_version = portable_version
        self._class_defs = dict()  # {factory_id:ClassDefinitionContext}

    def get_class_version(self, factory_id, class_id):
        return self._get_class_def_context(factory_id).get_class_version(class_id)

    def set_class_version(self, factory_id, class_id, version):
        self._get_class_def_context(factory_id).set_class_version(class_id, version)

    def lookup_class_definition(self, factory_id, class_id, version):
        return self._get_class_def_context(factory_id).lookup(class_id, version)

    def read_class_definition(self, data_in, factory_id, class_id, version):
        register = True
        builder = ClassDefinitionBuilder(factory_id, class_id, version)

        # final position after portable is read
        data_in.read_int()

        # field count
        field_count = data_in.read_int()
        offset = data_in.position()
        for i in range(0, field_count):
            pos = data_in.read_int(offset + i * bits.INT_SIZE_IN_BYTES)
            data_in.set_position(pos)

            _len = data_in.read_short()
            field_name = bytearray(_len)
            data_in.read_into(field_name)

            field_type = data_in.read_byte()

            field_factory_id = 0
            field_class_id = 0
            field_version = version
            if field_type == FieldType.PORTABLE:
                # is null
                if data_in.read_boolean():
                    register = False
                field_factory_id = data_in.read_int()
                field_class_id = data_in.read_int()

                # TODO: what there's a null inner Portable field
                if register:
                    field_version = data_in.read_int()
                    self.read_class_definition(
                        data_in, field_factory_id, field_class_id, field_version
                    )
            elif field_type == FieldType.PORTABLE_ARRAY:
                k = data_in.read_int()
                field_factory_id = data_in.read_int()
                field_class_id = data_in.read_int()

                # TODO: what there's a null inner Portable field
                if k > 0:
                    p = data_in.read_int()
                    data_in.set_position(p)
                    field_version = data_in.read_int()
                    self.read_class_definition(
                        data_in, field_factory_id, field_class_id, field_version
                    )
                else:
                    register = False
            builder.add_field_def(
                FieldDefinition(
                    i,
                    field_name.decode("ascii"),
                    field_type,
                    field_version,
                    field_factory_id,
                    field_class_id,
                )
            )
        class_def = builder.build()
        if register:
            class_def = self.register_class_definition(class_def)
        return class_def

    def register_class_definition(self, class_definition):
        return self._get_class_def_context(class_definition.factory_id).register(class_definition)

    def lookup_or_register_class_definition(self, portable):
        fid = portable.get_factory_id()
        cid = portable.get_class_id()
        portable_version = util.get_portable_version(portable, self.portable_version)
        class_def = self.lookup_class_definition(fid, cid, portable_version)

        if class_def is None:
            writer = ClassDefinitionWriter(self, fid, cid, portable_version)
            portable.write_portable(writer)
            class_def = writer.register_and_get()

        return class_def

    def _get_class_def_context(self, factory_id):
        if factory_id not in self._class_defs:
            self._class_defs[factory_id] = ClassDefinitionContext(factory_id, self.portable_version)
        return self._class_defs[factory_id]


class ClassDefinitionContext(object):
    def __init__(self, factory_id, portable_version):
        self._factory_id = factory_id
        self._portable_version = portable_version
        self._versioned_definitions = {}  # (class_id, version) : ClassDefinition
        self._current_class_versions = {}  # class_id:version
        self._lock = threading.RLock()

    def get_class_version(self, class_id):
        return self._current_class_versions.get(class_id, -1)

    def set_class_version(self, class_id, version):
        try:
            current_version = self._current_class_versions[class_id]
            if current_version != version:
                raise ValueError("Class-id: %s is already registered!" % class_id)
        except KeyError:
            self._current_class_versions[class_id] = version

    def lookup(self, class_id, version):
        return self._versioned_definitions.get((class_id, version), None)

    def register(self, class_def):
        with self._lock:
            if class_def is None:
                return None
            if class_def.factory_id != self._factory_id:
                raise HazelcastSerializationError(
                    "Invalid factory-id! %s -> %s" % (self._factory_id, class_def)
                )
            if isinstance(class_def, ClassDefinition):
                class_def.set_version_if_not_set(self._portable_version)
            combined_key = (class_def.class_id, class_def.version)
            if combined_key not in self._versioned_definitions:
                self._versioned_definitions[combined_key] = class_def
                return class_def
            current_class_def = self._versioned_definitions[combined_key]
            if isinstance(current_class_def, ClassDefinition):
                if current_class_def != class_def:
                    raise HazelcastSerializationError(
                        "Incompatible class-definitions with same class-id: %s vs %s"
                        % (class_def, current_class_def)
                    )
                return current_class_def
            self._versioned_definitions[combined_key] = class_def
            return class_def



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec

# hex: 0x010A00
_REQUEST_MESSAGE_TYPE = 68096
# hex: 0x010A01
_RESPONSE_MESSAGE_TYPE = 68097

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(name):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, False)



import zlib

from hazelcast import six
from hazelcast.serialization import (
    BYTE_SIZE_IN_BYTES,
    CHAR_SIZE_IN_BYTES,
    BE_UINT16,
    DOUBLE_SIZE_IN_BYTES,
    BE_DOUBLE,
    INT_SIZE_IN_BYTES,
    BE_INT,
    LONG_SIZE_IN_BYTES,
    BE_LONG,
    BE_UINT8,
)

_MASK_PREFIX = 1
_MASK_METRIC = 1 << 1
_MASK_DISCRIMINATOR = 1 << 2
_MASK_DISCRIMINATOR_VALUE = 1 << 3
_MASK_UNIT = 1 << 4
_MASK_EXCLUDED_TARGETS = 1 << 5
_MASK_TAG_COUNT = 1 << 6

_NULL_DICTIONARY_ID = -1
_NULL_UNIT = -1
_MAX_WORD_LENGTH = 255

_BITS_IN_BYTE = 8
_BYTE_MASK = 0xFF
_BINARY_FORMAT_VERSION = 1
_SIZE_VERSION = 2
_SIZE_DICTIONARY_BLOB = 4
_SIZE_COUNT_METRICS = 4

_OUTPUT_BUFFER_INITIAL_SIZE = 1024
_OUTPUT_BUFFER_GROW_FACTOR = 1.2


class MetricDescriptor(object):
    """Describes a metric to be sent to the members.

    It is a simplified version of the Java's MetricDescriptorImpl, sufficient
    for the needs of the Python client.
    """

    __slots__ = ("prefix", "metric", "discriminator", "discriminator_value", "unit")

    def __init__(
        self, metric, prefix=None, discriminator=None, discriminator_value=None, unit=None
    ):
        self.metric = metric
        """str: Name of the metric."""

        self.prefix = prefix
        """str: Prefix of the metric."""

        self.discriminator = discriminator
        """str: Discriminator for the metrics that have the same name and prefix."""

        self.discriminator_value = discriminator_value
        """str: Tag of the discriminator."""

        self.unit = unit
        """ProbeUnit: Unit of the metric."""


class ProbeUnit(object):
    """Measurement unit of a probe.

    The values of the constants below should be in sync with the ProbeUnit
    enum in Java.
    """

    BYTES = 0
    """Size or counter represented in bytes."""

    MS = 1
    """Timestamp or duration represented in milliseconds."""

    NS = 2
    """Timestamp or duration represented in nanoseconds."""

    PERCENT = 3
    """An integer mostly in range 0..100 or a double mostly in range 0..1."""

    COUNT = 4
    """Number of items: size, counter..."""

    BOOLEAN = 5
    """0 or 1."""

    ENUM = 6
    """0..n, ordinal of an enum."""

    # New enum values should be converted to a tag to ensure backward
    # compatibility during compressing the metrics. see ProbeUnit#newUnit
    # handling in Java for that. We don't implement this functionality yet in
    # the Python client, as we don't use such enum members.


class ValueType(object):
    """Type of the metric values.

    The values of the constants below should be in sync ValueType enum in Java.
    """

    LONG = 0
    DOUBLE = 1


class MetricsCompressor(object):
    """Compresses metrics into a ``bytearray`` blob.

    The compressor uses dictionary based delta compression and deflates
    the resulting ``bytearray`` by using ``zlib.compress()``. This
    compressor doesn't use the textual representation of the
    :class:`MetricDescriptor` hence it is agnostic to the order of the
    tags in that representation.

    Adding metrics by calling :func:`add_long` or :func:`add_double`
    builds a dictionary by mapping all words found in the passed
    :class:`MetricDescriptor`s to ``int``s and these ``int``s will be
    written to the resulting ``bytearray`` blob. Before these ``int``s
    are written, the current :class:`MetricDescriptor` is compared to the
    previous one and only the fields of the descriptor that are different
    from the previous will be written to the metrics blob.

    When the blob is retrieved from this compressor (when the metrics
    collection cycle finishes) the dictionary is stored in the dictionary
    blob. The compressor iterates over the words stored in the dictionary
    in ascending order and writes them to the blob by skipping the first
    N characters that are equal to the first N character of the previously
    written word, hence using delta compression here too.

    After both the metrics and the dictionary blob is constructed, they
    are copied into a final blob in the following structure:

    +--------------------------------+--------------------+
    | Compressor version             |   2 bytes (short)  |
    +--------------------------------+--------------------+
    | Size of dictionary blob        |   4 bytes (int)    |
    +--------------------------------+--------------------+
    | Dictionary blob                |   variable size    |
    +--------------------------------+--------------------+
    | Number of metrics in the blob  |   4 bytes (int)    |
    +--------------------------------+--------------------+
    | Metrics blob                   |   variable size    |
    +--------------------------------+--------------------+
    """

    __slots__ = ("_metrics_buf", "_dict_buf", "_metrics_dict", "_metrics_count", "_last_descriptor")

    def __init__(self):
        self._metrics_buf = _OutputBuffer()
        self._dict_buf = _OutputBuffer()
        self._metrics_dict = _MetricsDictionary()
        self._metrics_count = 0
        self._last_descriptor = None

    def add_long(self, descriptor, value):
        self._write_descriptor(descriptor)
        self._metrics_buf.write_byte(ValueType.LONG)
        self._metrics_buf.write_long(value)

    def add_double(self, descriptor, value):
        self._write_descriptor(descriptor)
        self._metrics_buf.write_byte(ValueType.DOUBLE)
        self._metrics_buf.write_double(value)

    def generate_blob(self):
        self._write_metrics_dict()
        metrics_buf = self._metrics_buf.compress()
        dict_buf = self._dict_buf.compress()

        complete_size = (
            _SIZE_VERSION
            + _SIZE_DICTIONARY_BLOB
            + len(dict_buf)
            + _SIZE_COUNT_METRICS
            + len(metrics_buf)
        )
        final_buf = _OutputBuffer(complete_size)
        final_buf.write_byte((_BINARY_FORMAT_VERSION >> _BITS_IN_BYTE) & _BYTE_MASK)
        final_buf.write_byte(_BINARY_FORMAT_VERSION & _BYTE_MASK)
        final_buf.write_int(len(dict_buf))
        final_buf.write_bytearray(dict_buf)
        final_buf.write_int(self._metrics_count)
        final_buf.write_bytearray(metrics_buf)
        return final_buf.to_bytearray()

    def _write_descriptor(self, descriptor):
        mask = self._calculate_descriptor_mask(descriptor)
        self._metrics_buf.write_byte(mask)

        # Only write to metrics buffer if the values below are
        # different than the values of the previous descriptor
        if mask & _MASK_PREFIX == 0:
            self._metrics_buf.write_int(self._get_dict_id(descriptor.prefix))

        if mask & _MASK_METRIC == 0:
            self._metrics_buf.write_int(self._get_dict_id(descriptor.metric))

        if mask & _MASK_DISCRIMINATOR == 0:
            self._metrics_buf.write_int(self._get_dict_id(descriptor.discriminator))

        if mask & _MASK_DISCRIMINATOR_VALUE == 0:
            self._metrics_buf.write_int(self._get_dict_id(descriptor.discriminator_value))

        if mask & _MASK_UNIT == 0:
            if descriptor.unit is None:
                self._metrics_buf.write_byte(_NULL_UNIT)
            else:
                self._metrics_buf.write_byte(descriptor.unit)

        # include excludedTargets and tags bytes for compatibility purposes
        if mask & _MASK_EXCLUDED_TARGETS == 0:
            self._metrics_buf.write_byte(0)

        if mask & _MASK_TAG_COUNT == 0:
            self._metrics_buf.write_byte(0)

        self._metrics_count += 1
        self._last_descriptor = descriptor

    def _calculate_descriptor_mask(self, descriptor):
        mask = 0
        if not self._last_descriptor:
            return mask

        last_descriptor = self._last_descriptor

        if descriptor.prefix == last_descriptor.prefix:
            mask |= _MASK_PREFIX

        if descriptor.metric == last_descriptor.metric:
            mask |= _MASK_METRIC

        if descriptor.discriminator == last_descriptor.discriminator:
            mask |= _MASK_DISCRIMINATOR

        if descriptor.discriminator_value == last_descriptor.discriminator_value:
            mask |= _MASK_DISCRIMINATOR_VALUE

        if descriptor.unit == last_descriptor.unit:
            mask |= _MASK_UNIT

        # include excludedTargets and tags bits for compatibility purposes
        mask |= _MASK_EXCLUDED_TARGETS
        mask |= _MASK_TAG_COUNT

        return mask

    def _get_dict_id(self, word):
        if not word:
            return _NULL_DICTIONARY_ID

        return self._metrics_dict.get_dict_id(word)

    def _write_metrics_dict(self):
        words = self._metrics_dict.get_words()
        self._dict_buf.write_int(len(words))

        last_word_text = ""
        for word in words:
            word_text = word.word
            max_common_len = min(len(last_word_text), len(word_text))
            common_len = 0
            while (
                common_len < max_common_len and word_text[common_len] == last_word_text[common_len]
            ):
                common_len += 1

            diff_len = len(word_text) - common_len

            self._dict_buf.write_int(word.dict_id)
            self._dict_buf.write_byte(common_len)
            self._dict_buf.write_byte(diff_len)
            for i in range(common_len, len(word_text)):
                self._dict_buf.write_char(word_text[i])

            last_word_text = word_text


class _OutputBuffer(object):
    __slots__ = ("_buf", "_pos")

    def __init__(self, size=None):
        self._buf = bytearray(size or _OUTPUT_BUFFER_INITIAL_SIZE)
        self._pos = 0

    def to_bytearray(self):
        if self._pos == len(self._buf):
            return self._buf

        return self._buf[: self._pos]

    def write_bytearray(self, buf):
        n = len(buf)
        self._ensure_available(n)
        self._buf[self._pos : self._pos + n] = buf
        self._pos += n

    def write_byte(self, value):
        self._ensure_available(BYTE_SIZE_IN_BYTES)
        BE_UINT8.pack_into(self._buf, self._pos, value & _BYTE_MASK)
        self._pos += BYTE_SIZE_IN_BYTES

    def write_char(self, value):
        self._ensure_available(CHAR_SIZE_IN_BYTES)
        BE_UINT16.pack_into(self._buf, self._pos, ord(value))
        self._pos += CHAR_SIZE_IN_BYTES

    def write_double(self, value):
        self._ensure_available(DOUBLE_SIZE_IN_BYTES)
        BE_DOUBLE.pack_into(self._buf, self._pos, value)
        self._pos += DOUBLE_SIZE_IN_BYTES

    def write_int(self, value):
        self._ensure_available(INT_SIZE_IN_BYTES)
        BE_INT.pack_into(self._buf, self._pos, value)
        self._pos += INT_SIZE_IN_BYTES

    def write_long(self, value):
        self._ensure_available(LONG_SIZE_IN_BYTES)
        BE_LONG.pack_into(self._buf, self._pos, value)
        self._pos += LONG_SIZE_IN_BYTES

    def compress(self):
        buf = self.to_bytearray()
        if six.PY2:
            # Python2 expects this to be read-only buffer
            buf = bytes(buf)

        # Set level to 1 for best speed (less CPU overhead)
        return zlib.compress(buf, 1)

    def _ensure_available(self, size):
        if self._available() < size:
            # Grow memory more than needed
            new_size = int((self._pos + size) * _OUTPUT_BUFFER_GROW_FACTOR)
            if new_size % 2 != 0:
                new_size += 1

            new_buf = bytearray(new_size)
            new_buf[: self._pos] = self._buf
            self._buf = new_buf

    def _available(self):
        return len(self._buf) - self._pos


class _Word(object):
    __slots__ = ("word", "dict_id")

    def __init__(self, word, dict_id):
        self.word = word
        self.dict_id = dict_id


class _MetricsDictionary(object):
    """Stores word -> id mappings.

    Used by :class:`MetricsCompressor`'s dictionary-based algorithm.
    """

    __slots__ = ("_words",)

    def __init__(self):
        self._words = {}

    def get_dict_id(self, word_text):
        """Returns the dictionary id for the given word.

        If the word is not yet stored in the dictionary, the word gets stored
        and a newly assigned id is returned.

        Args:
            word_text (str): Textual representation of a word.

        Returns:
            int: The dictionary id.

        Raises:
            ValueError: If the length of the word cannot fit into unsigned
                byte range.
        """
        if len(word_text) > _MAX_WORD_LENGTH:
            raise ValueError(
                "Too long value in the metric descriptor found, maximum is %s: %s"
                % (_MAX_WORD_LENGTH, word_text)
            )

        word = self._words.get(word_text, None)
        if word:
            word_id = word.dict_id
        else:
            word_id = len(self._words)
            self._words[word_text] = _Word(word_text, word_id)

        return word_id

    def get_words(self):
        """Returns all stored word<->id mappings ordered by word.

        Returns:
            list[_Word]: Words.
        """
        words = self._words.values()
        return sorted(words, key=lambda w: w.word)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec
from hazelcast.protocol.builtin import StringCodec

# hex: 0x0C0200
_REQUEST_MESSAGE_TYPE = 786944
# hex: 0x0C0201
_RESPONSE_MESSAGE_TYPE = 786945

_REQUEST_SESSION_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_THREAD_ID_OFFSET = _REQUEST_SESSION_ID_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_INVOCATION_UID_OFFSET = _REQUEST_THREAD_ID_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_PERMITS_OFFSET = _REQUEST_INVOCATION_UID_OFFSET + UUID_SIZE_IN_BYTES
_REQUEST_TIMEOUT_MS_OFFSET = _REQUEST_PERMITS_OFFSET + INT_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_TIMEOUT_MS_OFFSET + LONG_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE


def encode_request(group_id, name, session_id, thread_id, invocation_uid, permits, timeout_ms):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_SESSION_ID_OFFSET, session_id)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_THREAD_ID_OFFSET, thread_id)
    FixSizedTypesCodec.encode_uuid(buf, _REQUEST_INVOCATION_UID_OFFSET, invocation_uid)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_PERMITS_OFFSET, permits)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_TIMEOUT_MS_OFFSET, timeout_ms)
    RaftGroupIdCodec.encode(buf, group_id)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_boolean(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE, EVENT_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec

# hex: 0x000900
_REQUEST_MESSAGE_TYPE = 2304
# hex: 0x000901
_RESPONSE_MESSAGE_TYPE = 2305
# hex: 0x000902
_EVENT_DISTRIBUTED_OBJECT_MESSAGE_TYPE = 2306

_REQUEST_LOCAL_ONLY_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_LOCAL_ONLY_OFFSET + BOOLEAN_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE
_EVENT_DISTRIBUTED_OBJECT_SOURCE_OFFSET = EVENT_HEADER_SIZE


def encode_request(local_only):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE, True)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_LOCAL_ONLY_OFFSET, local_only)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)


def handle(msg, handle_distributed_object_event=None):
    message_type = msg.get_message_type()
    if message_type == _EVENT_DISTRIBUTED_OBJECT_MESSAGE_TYPE and handle_distributed_object_event is not None:
        initial_frame = msg.next_frame()
        source = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _EVENT_DISTRIBUTED_OBJECT_SOURCE_OFFSET)
        name = StringCodec.decode(msg)
        service_name = StringCodec.decode(msg)
        event_type = StringCodec.decode(msg)
        handle_distributed_object_event(name, service_name, event_type, source)
        return



from hazelcast.invocation import Invocation
from hazelcast.protocol.codec import client_create_proxy_codec, client_destroy_proxy_codec
from hazelcast.proxy.executor import Executor
from hazelcast.proxy.list import List
from hazelcast.proxy.map import create_map_proxy
from hazelcast.proxy.multi_map import MultiMap
from hazelcast.proxy.queue import Queue
from hazelcast.proxy.reliable_topic import ReliableTopic
from hazelcast.proxy.replicated_map import ReplicatedMap
from hazelcast.proxy.ringbuffer import Ringbuffer
from hazelcast.proxy.set import Set
from hazelcast.proxy.topic import Topic
from hazelcast.proxy.pn_counter import PNCounter
from hazelcast.proxy.flake_id_generator import FlakeIdGenerator
from hazelcast.util import to_list

EXECUTOR_SERVICE = "hz:impl:executorService"
LIST_SERVICE = "hz:impl:listService"
MULTI_MAP_SERVICE = "hz:impl:multiMapService"
MAP_SERVICE = "hz:impl:mapService"
RELIABLE_TOPIC_SERVICE = "hz:impl:reliableTopicService"
REPLICATED_MAP_SERVICE = "hz:impl:replicatedMapService"
RINGBUFFER_SERVICE = "hz:impl:ringbufferService"
SET_SERVICE = "hz:impl:setService"
QUEUE_SERVICE = "hz:impl:queueService"
TOPIC_SERVICE = "hz:impl:topicService"
PN_COUNTER_SERVICE = "hz:impl:PNCounterService"
FLAKE_ID_GENERATOR_SERVICE = "hz:impl:flakeIdGeneratorService"

_proxy_init = {
    EXECUTOR_SERVICE: Executor,
    LIST_SERVICE: List,
    MAP_SERVICE: create_map_proxy,
    MULTI_MAP_SERVICE: MultiMap,
    QUEUE_SERVICE: Queue,
    RELIABLE_TOPIC_SERVICE: ReliableTopic,
    REPLICATED_MAP_SERVICE: ReplicatedMap,
    RINGBUFFER_SERVICE: Ringbuffer,
    SET_SERVICE: Set,
    TOPIC_SERVICE: Topic,
    PN_COUNTER_SERVICE: PNCounter,
    FLAKE_ID_GENERATOR_SERVICE: FlakeIdGenerator,
}


class ProxyManager(object):
    def __init__(self, context):
        self._context = context
        self._proxies = {}

    def get_or_create(self, service_name, name, create_on_remote=True):
        ns = (service_name, name)
        if ns in self._proxies:
            return self._proxies[ns]

        proxy = self._create_proxy(service_name, name, create_on_remote)
        self._proxies[ns] = proxy
        return proxy

    def _create_proxy(self, service_name, name, create_on_remote):
        if create_on_remote:
            request = client_create_proxy_codec.encode_request(name, service_name)
            invocation = Invocation(request)
            invocation_service = self._context.invocation_service
            invocation_service.invoke(invocation)
            invocation.future.result()

        return _proxy_init[service_name](service_name, name, self._context)

    def destroy_proxy(self, service_name, name, destroy_on_remote=True):
        ns = (service_name, name)
        try:
            self._proxies.pop(ns)
            if destroy_on_remote:
                request = client_destroy_proxy_codec.encode_request(name, service_name)
                invocation = Invocation(request)
                invocation_service = self._context.invocation_service
                invocation_service.invoke(invocation)
                invocation.future.result()
            return True
        except KeyError:
            return False

    def get_distributed_objects(self):
        return to_list(self._proxies.values())



from uuid import uuid4
from hazelcast import future
from hazelcast.protocol.codec import (
    executor_service_shutdown_codec,
    executor_service_is_shutdown_codec,
    executor_service_submit_to_partition_codec,
    executor_service_submit_to_member_codec,
)
from hazelcast.proxy.base import Proxy
from hazelcast.util import check_not_none


class Executor(Proxy):
    """An object that executes submitted executable tasks."""

    def execute_on_key_owner(self, key, task):
        """Executes a task on the owner of the specified key.

        Args:
          key: The specified key.
          task: A task executed on the owner of the specified key.

        Returns:
          hazelcast.future.Future: future representing pending completion of the task.
        """
        check_not_none(key, "key can't be None")
        check_not_none(task, "task can't be None")

        def handler(message):
            return self._to_object(
                executor_service_submit_to_partition_codec.decode_response(message)
            )

        key_data = self._to_data(key)
        task_data = self._to_data(task)

        partition_id = self._context.partition_service.get_partition_id(key_data)
        uuid = uuid4()
        request = executor_service_submit_to_partition_codec.encode_request(
            self.name, uuid, task_data
        )
        return self._invoke_on_partition(request, partition_id, handler)

    def execute_on_member(self, member, task):
        """Executes a task on the specified member.

        Args:
          member (hazelcast.core.MemberInfo): The specified member.
          task: The task executed on the specified member.

        Returns:
          hazelcast.future.Future: Future representing pending completion of the task.
        """
        check_not_none(task, "task can't be None")
        task_data = self._to_data(task)
        uuid = uuid4()
        return self._execute_on_member(uuid, task_data, member.uuid)

    def execute_on_members(self, members, task):
        """Executes a task on each of the specified members.

        Args:
          members (list[hazelcast.core.MemberInfo]): The specified members.
          task: The task executed on the specified members.

        Returns:
          list[hazelcast.future.Future]: Futures representing pending completion of the task on each member.

        """
        task_data = self._to_data(task)
        futures = []
        uuid = uuid4()
        for member in members:
            f = self._execute_on_member(uuid, task_data, member.uuid)
            futures.append(f)
        return future.combine_futures(futures)

    def execute_on_all_members(self, task):
        """Executes a task on all of the known cluster members.

        Args:
          task: The task executed on the all of the members.

        Returns:
          list[hazelcast.future.Future]: Futures representing pending completion of the task on each member.
        """
        return self.execute_on_members(self._context.cluster_service.get_members(), task)

    def is_shutdown(self):
        """Determines whether this executor has been shutdown or not.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the executor has been shutdown, ``False`` otherwise.
        """
        request = executor_service_is_shutdown_codec.encode_request(self.name)
        return self._invoke(request, executor_service_is_shutdown_codec.decode_response)

    def shutdown(self):
        """Initiates a shutdown process which works orderly. Tasks that were submitted
        before shutdown are executed but new task will not be accepted.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = executor_service_shutdown_codec.encode_request(self.name)
        return self._invoke(request)

    def _execute_on_member(self, uuid, task_data, member_uuid):
        def handler(message):
            return self._to_object(executor_service_submit_to_member_codec.decode_response(message))

        request = executor_service_submit_to_member_codec.encode_request(
            self.name, uuid, task_data, member_uuid
        )
        return self._invoke_on_target(request, member_uuid, handler)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x011100
_REQUEST_MESSAGE_TYPE = 69888
# hex: 0x011101
_RESPONSE_MESSAGE_TYPE = 69889

_REQUEST_THREAD_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_LEASE_OFFSET = _REQUEST_THREAD_ID_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_TIMEOUT_OFFSET = _REQUEST_LEASE_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_REFERENCE_ID_OFFSET = _REQUEST_TIMEOUT_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_REFERENCE_ID_OFFSET + LONG_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE


def encode_request(name, key, thread_id, lease, timeout, reference_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_THREAD_ID_OFFSET, thread_id)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_LEASE_OFFSET, lease)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_TIMEOUT_OFFSET, timeout)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_REFERENCE_ID_OFFSET, reference_id)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, key, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_boolean(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)



import logging
import os

from hazelcast.core import CLIENT_TYPE
from hazelcast.invocation import Invocation
from hazelcast.metrics import MetricsCompressor, MetricDescriptor, ValueType, ProbeUnit
from hazelcast.protocol.codec import client_statistics_codec
from hazelcast.util import current_time_in_millis, to_millis, to_nanos, current_time
from hazelcast import six, __version__

try:
    import psutil

    _PSUTIL_ENABLED = True
except ImportError:
    _PSUTIL_ENABLED = False

_logger = logging.getLogger(__name__)

_NEAR_CACHE_CATEGORY_PREFIX = "nc."
_ATTRIBUTE_SEPARATOR = ","
_KEY_VALUE_SEPARATOR = "="
_EMPTY_ATTRIBUTE_VALUE = ""

_NEAR_CACHE_DESCRIPTOR_PREFIX = "nearcache"
_NEAR_CACHE_DESCRIPTOR_DISCRIMINATOR = "name"


class Statistics(object):
    def __init__(
        self, client, config, reactor, connection_manager, invocation_service, near_cache_manager
    ):
        self._client = client
        self._reactor = reactor
        self._connection_manager = connection_manager
        self._invocation_service = invocation_service
        self._near_cache_manager = near_cache_manager
        self._enabled = config.statistics_enabled
        self._period = config.statistics_period
        self._statistics_timer = None
        self._registered_system_gauges = {}
        self._registered_process_gauges = {}

    def start(self):
        if not self._enabled:
            return

        self._register_gauges()

        def _statistics_task():
            if not self._client.lifecycle_service.is_running():
                return

            try:
                self._collect_and_send_stats()
            finally:
                self._statistics_timer = self._reactor.add_timer(self._period, _statistics_task)

        self._statistics_timer = self._reactor.add_timer(self._period, _statistics_task)

        _logger.info("Client statistics enabled with the period of %s seconds.", self._period)

    def shutdown(self):
        if self._statistics_timer:
            self._statistics_timer.cancel()

    def _register_gauges(self):
        if not _PSUTIL_ENABLED:
            _logger.warning(
                "Statistics collection is enabled, but psutil is not found. "
                "Runtime and system related metrics will not be collected."
            )
            return

        self._register_system_gauge(
            "os.totalPhysicalMemorySize",
            lambda: psutil.virtual_memory().total,
        )
        self._register_system_gauge(
            "os.freePhysicalMemorySize",
            lambda: psutil.virtual_memory().free,
        )
        self._register_system_gauge(
            "os.committedVirtualMemorySize",
            lambda: psutil.virtual_memory().used,
        )
        self._register_system_gauge(
            "os.totalSwapSpaceSize",
            lambda: psutil.swap_memory().total,
        )
        self._register_system_gauge(
            "os.freeSwapSpaceSize",
            lambda: psutil.swap_memory().free,
        )
        self._register_system_gauge(
            "os.systemLoadAverage",
            lambda: os.getloadavg()[0],
            ValueType.DOUBLE,
        )
        self._register_system_gauge(
            "runtime.availableProcessors",
            lambda: psutil.cpu_count(),
        )

        self._register_process_gauge(
            "runtime.usedMemory",
            lambda p: p.memory_info().rss,
        )
        self._register_process_gauge(
            "os.openFileDescriptorCount",
            lambda p: p.num_fds(),
        )
        self._register_process_gauge(
            "os.maxFileDescriptorCount",
            lambda p: p.rlimit(psutil.RLIMIT_NOFILE)[1],
        )
        self._register_process_gauge(
            "os.processCpuTime",
            lambda p: to_nanos(sum(p.cpu_times())),
        )
        self._register_process_gauge(
            "runtime.uptime",
            lambda p: to_millis(current_time() - p.create_time()),
        )

    def _register_system_gauge(self, gauge_name, gauge_fn, value_type=ValueType.LONG):
        # Try a gauge function read, we will register it if it succeeds.
        try:
            gauge_fn()
            self._registered_system_gauges[gauge_name] = (gauge_fn, value_type)
        except Exception as e:
            _logger.debug(
                "Unable to register the system related gauge %s. Error: %s" % (gauge_name, e)
            )

    def _register_process_gauge(self, gauge_name, gauge_fn, value_type=ValueType.LONG):
        # Try a gauge function read, we will register it if it succeeds.
        try:
            process = psutil.Process()
            gauge_fn(process)
            self._registered_process_gauges[gauge_name] = (gauge_fn, value_type)
        except Exception as e:
            _logger.debug(
                "Unable to register the process related gauge %s. Error: %s" % (gauge_name, e)
            )

    def _collect_and_send_stats(self):
        connection = self._connection_manager.get_random_connection()
        if not connection:
            _logger.debug("Cannot send client statistics to the server. No connection found.")
            return

        collection_timestamp = current_time_in_millis()
        attributes = []
        compressor = MetricsCompressor()

        self._add_client_attributes(attributes, connection)
        self._add_near_cache_metrics(attributes, compressor)
        self._add_system_and_process_metrics(attributes, compressor)
        self._send_stats(
            collection_timestamp, "".join(attributes), compressor.generate_blob(), connection
        )

    def _send_stats(self, collection_timestamp, attributes, metrics_blob, connection):
        request = client_statistics_codec.encode_request(
            collection_timestamp, attributes, metrics_blob
        )
        invocation = Invocation(request, connection=connection)
        self._invocation_service.invoke(invocation)

    def _add_system_and_process_metrics(self, attributes, compressor):
        if not _PSUTIL_ENABLED:
            # Nothing to do if psutil is not found
            return

        for gauge_name, (gauge_fn, value_type) in six.iteritems(self._registered_system_gauges):
            try:
                value = gauge_fn()
                self._add_system_or_process_metric(
                    attributes, compressor, gauge_name, value, value_type
                )
            except:
                _logger.exception("Error while collecting '%s'." % gauge_name)

        if not self._registered_process_gauges:
            # Do not create the process object if no process-related
            # metric is registered.
            return

        process = psutil.Process()
        for gauge_name, (gauge_fn, value_type) in six.iteritems(self._registered_process_gauges):
            try:
                value = gauge_fn(process)
                self._add_system_or_process_metric(
                    attributes, compressor, gauge_name, value, value_type
                )
            except:
                _logger.exception("Error while collecting '%s'." % gauge_name)

    def _add_system_or_process_metric(self, attributes, compressor, gauge_name, value, value_type):
        # We don't have any metrics that do not have prefix.
        # Necessary care must be taken when we will send simple
        # named metrics.
        prefix, metric_name = gauge_name.rsplit(".", 1)
        descriptor = MetricDescriptor(metric=metric_name, prefix=prefix)
        self._add_metric(compressor, descriptor, value, value_type)
        self._add_attribute(attributes, gauge_name, value)

    def _add_client_attributes(self, attributes, connection):
        self._add_attribute(attributes, "lastStatisticsCollectionTime", current_time_in_millis())
        self._add_attribute(attributes, "enterprise", "false")
        self._add_attribute(attributes, "clientType", CLIENT_TYPE)
        self._add_attribute(attributes, "clientVersion", __version__)
        self._add_attribute(
            attributes, "clusterConnectionTimestamp", to_millis(connection.start_time)
        )

        local_address = connection.local_address
        local_address = str(local_address.host) + ":" + str(local_address.port)
        self._add_attribute(attributes, "clientAddress", local_address)
        self._add_attribute(attributes, "clientName", self._client.name)

    def _add_near_cache_metrics(self, attributes, compressor):
        for near_cache in self._near_cache_manager.list_near_caches():
            nc_name = near_cache.name
            nc_name_with_prefix = self._get_name_with_prefix(nc_name)
            nc_name_with_prefix.append(".")
            nc_name_with_prefix = "".join(nc_name_with_prefix)

            near_cache_stats = near_cache.get_statistics()
            self._add_near_cache_metric(
                attributes,
                compressor,
                "creationTime",
                to_millis(near_cache_stats["creation_time"]),
                ValueType.LONG,
                ProbeUnit.MS,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "evictions",
                near_cache_stats["evictions"],
                ValueType.LONG,
                ProbeUnit.COUNT,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "hits",
                near_cache_stats["hits"],
                ValueType.LONG,
                ProbeUnit.COUNT,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "misses",
                near_cache_stats["misses"],
                ValueType.LONG,
                ProbeUnit.COUNT,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "ownedEntryCount",
                near_cache_stats["owned_entry_count"],
                ValueType.LONG,
                ProbeUnit.COUNT,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "expirations",
                near_cache_stats["expirations"],
                ValueType.LONG,
                ProbeUnit.COUNT,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "invalidations",
                near_cache_stats["invalidations"],
                ValueType.LONG,
                ProbeUnit.COUNT,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "invalidationRequests",
                near_cache_stats["invalidation_requests"],
                ValueType.LONG,
                ProbeUnit.COUNT,
                nc_name,
                nc_name_with_prefix,
            )

            self._add_near_cache_metric(
                attributes,
                compressor,
                "ownedEntryMemoryCost",
                near_cache_stats["owned_entry_memory_cost"],
                ValueType.LONG,
                ProbeUnit.BYTES,
                nc_name,
                nc_name_with_prefix,
            )

    def _add_near_cache_metric(
        self, attributes, compressor, metric, value, value_type, unit, nc_name, nc_name_with_prefix
    ):
        descriptor = MetricDescriptor(
            metric=metric,
            prefix=_NEAR_CACHE_DESCRIPTOR_PREFIX,
            discriminator=_NEAR_CACHE_DESCRIPTOR_DISCRIMINATOR,
            discriminator_value=nc_name,
            unit=unit,
        )
        try:
            self._add_metric(compressor, descriptor, value, value_type)
            self._add_attribute(attributes, metric, value, nc_name_with_prefix)
        except:
            _logger.exception(
                "Error while collecting %s metric for near cache '%s'" % (metric, nc_name)
            )

    def _add_metric(self, compressor, descriptor, value, value_type):
        if value_type == ValueType.LONG:
            compressor.add_long(descriptor, value)
        elif value_type == ValueType.DOUBLE:
            compressor.add_double(descriptor, value)
        else:
            raise ValueError("Unexpected type: " + value_type)

    def _add_attribute(self, attributes, name, value, key_prefix=None):
        if len(attributes) != 0:
            attributes.append(_ATTRIBUTE_SEPARATOR)

        if key_prefix:
            attributes.append(key_prefix)

        attributes.append(name)
        attributes.append(_KEY_VALUE_SEPARATOR)
        attributes.append(str(value))

    def _get_name_with_prefix(self, name):
        return [_NEAR_CACHE_CATEGORY_PREFIX, self._escape_special_characters(name)]

    def _escape_special_characters(self, name):
        escaped_name = (
            name.replace("\\", "\\\\").replace(",", "\\,").replace(".", "\\.").replace("=", "\\=")
        )
        return escaped_name[1:] if name[0] == "/" else escaped_name



from hazelcast.serialization import LE_UINT
from hazelcast.six.moves import range


def murmur_hash3_x86_32(data):
    """murmur3 hash function to determine partition

    Args:
        data (bytearray or bytes): Input byte array

    Returns:
        int: Calculated hash value.
    """
    length = max(len(data) - 8, 0)  # Heap data overhead
    nblocks = length // 4

    h1 = 0x01000193

    c1 = 0xCC9E2D51
    c2 = 0x1B873593

    # body
    for block_start in range(0, nblocks * 4, 4):
        # ??? big endian?
        k1 = LE_UINT.unpack_from(data, block_start + 8)[0]

        k1 = c1 * k1 & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        k1 = (c2 * k1) & 0xFFFFFFFF

        h1 ^= k1
        h1 = (h1 << 13 | h1 >> 19) & 0xFFFFFFFF  # inlined _ROTL32
        h1 = (h1 * 5 + 0xE6546B64) & 0xFFFFFFFF

    # tail
    tail_index = nblocks * 4
    k1 = 0
    tail_size = length & 3

    # Offsets below are shifted according to heap data overhead
    if tail_size >= 3:
        k1 ^= data[tail_index + 10] << 16
    if tail_size >= 2:
        k1 ^= data[tail_index + 9] << 8
    if tail_size >= 1:
        k1 ^= data[tail_index + 8]

    if tail_size != 0:
        k1 = (k1 * c1) & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # _ROTL32
        k1 = (k1 * c2) & 0xFFFFFFFF
        h1 ^= k1

    h1 ^= length
    h1 ^= h1 >> 16
    h1 = (h1 * 0x85EBCA6B) & 0xFFFFFFFF
    h1 ^= h1 >> 13
    h1 = (h1 * 0xC2B2AE35) & 0xFFFFFFFF
    h1 ^= h1 >> 16
    return -(h1 & 0x80000000) | (h1 & 0x7FFFFFFF)


def hash_to_index(mm_hash, length):
    if mm_hash == 0x80000000:
        return 0
    else:
        return abs(mm_hash) % length



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import EntryListCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x012C00
_REQUEST_MESSAGE_TYPE = 76800
# hex: 0x012C01
_RESPONSE_MESSAGE_TYPE = 76801

_REQUEST_TRIGGER_MAP_LOADER_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_TRIGGER_MAP_LOADER_OFFSET + BOOLEAN_SIZE_IN_BYTES


def encode_request(name, entries, trigger_map_loader):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_TRIGGER_MAP_LOADER_OFFSET, trigger_map_loader)
    StringCodec.encode(buf, name)
    EntryListCodec.encode(buf, entries, DataCodec.encode, DataCodec.encode, True)
    return OutboundMessage(buf, False)



from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE, EVENT_HEADER_SIZE

# hex: 0x000F00
_REQUEST_MESSAGE_TYPE = 3840
# hex: 0x000F01
_RESPONSE_MESSAGE_TYPE = 3841
# hex: 0x000F02
_EVENT_BACKUP_MESSAGE_TYPE = 3842

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE
_EVENT_BACKUP_SOURCE_INVOCATION_CORRELATION_ID_OFFSET = EVENT_HEADER_SIZE


def encode_request():
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)


def handle(msg, handle_backup_event=None):
    message_type = msg.get_message_type()
    if message_type == _EVENT_BACKUP_MESSAGE_TYPE and handle_backup_event is not None:
        initial_frame = msg.next_frame()
        source_invocation_correlation_id = FixSizedTypesCodec.decode_long(initial_frame.buf, _EVENT_BACKUP_SOURCE_INVOCATION_CORRELATION_ID_OFFSET)
        handle_backup_event(source_invocation_correlation_id)
        return



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.codec.custom.distributed_object_info_codec import DistributedObjectInfoCodec

# hex: 0x000800
_REQUEST_MESSAGE_TYPE = 2048
# hex: 0x000801
_RESPONSE_MESSAGE_TYPE = 2049

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request():
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    msg.next_frame()
    return ListMultiFrameCodec.decode(msg, DistributedObjectInfoCodec.decode)



from hazelcast.future import make_blocking
from hazelcast.invocation import Invocation
from hazelcast.partition import string_partition_strategy
from hazelcast import six
from hazelcast.util import get_attr_name

MAX_SIZE = float("inf")


def _no_op_response_handler(_):
    return None


class Proxy(object):
    """Provides basic functionality for Hazelcast Proxies."""

    def __init__(self, service_name, name, context):
        self.service_name = service_name
        self.name = name
        self._context = context
        self._invocation_service = context.invocation_service
        self._partition_service = context.partition_service
        serialization_service = context.serialization_service
        self._to_object = serialization_service.to_object
        self._to_data = serialization_service.to_data
        listener_service = context.listener_service
        self._register_listener = listener_service.register_listener
        self._deregister_listener = listener_service.deregister_listener
        self._is_smart = context.config.smart_routing

    def destroy(self):
        """Destroys this proxy.

        Returns:
            bool: ``True`` if this proxy is destroyed successfully, ``False`` otherwise.
        """
        self._on_destroy()
        return self._context.proxy_manager.destroy_proxy(self.service_name, self.name)

    def _on_destroy(self):
        pass

    def __repr__(self):
        return '%s(name="%s")' % (type(self).__name__, self.name)

    def _invoke(self, request, response_handler=_no_op_response_handler):
        invocation = Invocation(request, response_handler=response_handler)
        self._invocation_service.invoke(invocation)
        return invocation.future

    def _invoke_on_target(self, request, uuid, response_handler=_no_op_response_handler):
        invocation = Invocation(request, uuid=uuid, response_handler=response_handler)
        self._invocation_service.invoke(invocation)
        return invocation.future

    def _invoke_on_key(self, request, key_data, response_handler=_no_op_response_handler):
        partition_id = self._partition_service.get_partition_id(key_data)
        invocation = Invocation(
            request, partition_id=partition_id, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future

    def _invoke_on_partition(self, request, partition_id, response_handler=_no_op_response_handler):
        invocation = Invocation(
            request, partition_id=partition_id, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future

    def blocking(self):
        """Returns a version of this proxy with only blocking method calls."""
        return make_blocking(self)


class PartitionSpecificProxy(Proxy):
    """Provides basic functionality for Partition Specific Proxies."""

    def __init__(self, service_name, name, context):
        super(PartitionSpecificProxy, self).__init__(service_name, name, context)
        partition_key = context.serialization_service.to_data(string_partition_strategy(self.name))
        self._partition_id = context.partition_service.get_partition_id(partition_key)

    def _invoke(self, request, response_handler=_no_op_response_handler):
        invocation = Invocation(
            request, partition_id=self._partition_id, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future


class TransactionalProxy(object):
    """Provides an interface for all transactional distributed objects."""

    def __init__(self, name, transaction, context):
        self.name = name
        self.transaction = transaction
        self._invocation_service = context.invocation_service
        serialization_service = context.serialization_service
        self._to_object = serialization_service.to_object
        self._to_data = serialization_service.to_data

    def _invoke(self, request, response_handler=_no_op_response_handler):
        invocation = Invocation(
            request, connection=self.transaction.connection, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future

    def __repr__(self):
        return '%s(name="%s")' % (type(self).__name__, self.name)


class ItemEventType(object):
    """Type of item events."""

    ADDED = 1
    """
    Fired when an item is added.
    """

    REMOVED = 2
    """
    Fired when an item is removed.
    """


class EntryEventType(object):
    """Type of entry event."""

    ADDED = 1
    """
    Fired if an entry is added.
    """

    REMOVED = 2
    """
    Fired if an entry is removed.
    """

    UPDATED = 4
    """
    Fired if an entry is updated.
    """

    EVICTED = 8
    """
    Fired if an entry is evicted.
    """

    EXPIRED = 16
    """
    Fired if an entry is expired.
    """

    EVICT_ALL = 32
    """
    Fired if all entries are evicted.
    """

    CLEAR_ALL = 64
    """
    Fired if all entries are cleared.
    """

    MERGED = 128
    """
    Fired if an entry is merged after a network partition.
    """

    INVALIDATION = 256
    """
    Fired if an entry is invalidated.
    """

    LOADED = 512
    """
    Fired if an entry is loaded.
    """


class ItemEvent(object):
    """Map Item event.

    Attributes:
        name (str): Name of the proxy that fired the event.
        event_type (ItemEventType): Type of the event.
        member (hazelcast.core.MemberInfo): Member that fired the event.
    """

    def __init__(self, name, item_data, event_type, member, to_object):
        self.name = name
        self._item_data = item_data
        self.event_type = event_type
        self.member = member
        self._to_object = to_object

    @property
    def item(self):
        """The item related to the event."""
        return self._to_object(self._item_data)


class EntryEvent(object):
    """Map Entry event.

    Attributes:
        event_type (EntryEventType): Type of the event.
        uuid (uuid.UUID): UUID of the member that fired the event.
        number_of_affected_entries (int): Number of affected entries by this event.
    """

    def __init__(
        self,
        to_object,
        key,
        value,
        old_value,
        merging_value,
        event_type,
        uuid,
        number_of_affected_entries,
    ):
        self._to_object = to_object
        self._key_data = key
        self._value_data = value
        self._old_value_data = old_value
        self._merging_value_data = merging_value
        self.event_type = event_type
        self.uuid = uuid
        self.number_of_affected_entries = number_of_affected_entries

    @property
    def key(self):
        """The key of this entry event."""
        return self._to_object(self._key_data)

    @property
    def old_value(self):
        """The old value of the entry event."""
        return self._to_object(self._old_value_data)

    @property
    def value(self):
        """The value of the entry event."""
        return self._to_object(self._value_data)

    @property
    def merging_value(self):
        """The incoming merging value of the entry event."""
        return self._to_object(self._merging_value_data)

    def __repr__(self):
        return (
            "EntryEvent(key=%s, value=%s, old_value=%s, merging_value=%s, event_type=%s, uuid=%s, "
            "number_of_affected_entries=%s)"
            % (
                self.key,
                self.value,
                self.old_value,
                self.merging_value,
                get_attr_name(EntryEventType, self.event_type),
                self.uuid,
                self.number_of_affected_entries,
            )
        )


_SENTINEL = object()


class TopicMessage(object):
    """Topic message."""

    __slots__ = ("_name", "_message_data", "_message", "_publish_time", "_member", "_to_object")

    def __init__(self, name, message_data, publish_time, member, to_object):
        self._name = name
        self._message_data = message_data
        self._message = _SENTINEL
        self._publish_time = publish_time
        self._member = member
        self._to_object = to_object

    @property
    def name(self):
        """str: Name of the proxy that fired the event."""
        return self._name

    @property
    def publish_time(self):
        """int: UNIX time that the event is published as seconds."""
        return self._publish_time

    @property
    def member(self):
        """hazelcast.core.MemberInfo: Member that fired the event."""
        return self._member

    @property
    def message(self):
        """The message sent to Topic."""
        if self._message is not _SENTINEL:
            return self._message

        self._message = self._to_object(self._message_data)
        return self._message

    def __repr__(self):
        return "TopicMessage(message=%s, publish_time=%s, topic_name=%s, publishing_member=%s)" % (
            self.message,
            self.publish_time,
            self.name,
            self.member,
        )


def get_entry_listener_flags(**kwargs):
    flags = 0
    for (key, value) in six.iteritems(kwargs):
        if value:
            flags |= getattr(EntryEventType, key)
    return flags



import json
import logging
import ssl

from hazelcast.errors import HazelcastCertificationError
from hazelcast.core import AddressHelper
from hazelcast.six.moves import http_client

_logger = logging.getLogger(__name__)


class HazelcastCloudAddressProvider(object):
    """Provides initial addresses for client to find and connect to a node
    and resolves private IP addresses of Hazelcast Cloud service.
    """

    def __init__(self, token, connection_timeout):
        self.cloud_discovery = HazelcastCloudDiscovery(token, connection_timeout)
        self._private_to_public = dict()

    def load_addresses(self):
        """Loads member addresses from Hazelcast Cloud endpoint.

        Returns:
            tuple[list[hazelcast.core.Address], list[hazelcast.core.Address]]: The possible member addresses
                as primary addresses to connect to.
        """
        try:
            nodes = self.cloud_discovery.discover_nodes()
            # Every private address is primary
            return list(nodes.keys()), []
        except Exception as e:
            _logger.warning("Failed to load addresses from Hazelcast Cloud: %s", e)
        return [], []

    def translate(self, address):
        """Translates the given address to another address specific to network or service.

        Args:
            address (hazelcast.core.Address): Private address to be translated

        Returns:
            hazelcast.core.Address: New address if given address is known, otherwise returns None
        """
        if address is None:
            return None

        public_address = self._private_to_public.get(address, None)
        if public_address:
            return public_address

        self.refresh()

        return self._private_to_public.get(address, None)

    def refresh(self):
        """Refreshes the internal lookup table if necessary."""
        try:
            self._private_to_public = self.cloud_discovery.discover_nodes()
        except Exception as e:
            _logger.warning("Failed to load addresses from Hazelcast.cloud: %s", e)


class HazelcastCloudDiscovery(object):
    """Discovery service that discover nodes via Hazelcast.cloud
    https://coordinator.hazelcast.cloud/cluster/discovery?token=<TOKEN>
    """

    _CLOUD_URL_BASE = "coordinator.hazelcast.cloud"
    _CLOUD_URL_PATH = "/cluster/discovery?token="
    _PRIVATE_ADDRESS_PROPERTY = "private-address"
    _PUBLIC_ADDRESS_PROPERTY = "public-address"

    def __init__(self, token, connection_timeout):
        self._url = self._CLOUD_URL_PATH + token
        self._connection_timeout = connection_timeout
        # Default context operates only on TLSv1+, checks certificates,hostname and validity
        self._ctx = ssl.create_default_context()

    def discover_nodes(self):
        """Discovers nodes from Hazelcast.cloud.

        Returns:
            dict[hazelcast.core.Address, hazelcast.core.Address]: Dictionary that maps private
                addresses to public addresses.
        """
        try:
            https_connection = http_client.HTTPSConnection(
                host=self._CLOUD_URL_BASE, timeout=self._connection_timeout, context=self._ctx
            )
            https_connection.request(
                method="GET", url=self._url, headers={"Accept-Charset": "UTF-8"}
            )
            https_response = https_connection.getresponse()
        except ssl.SSLError as err:
            raise HazelcastCertificationError(str(err))
        self._check_error(https_response)
        return self._parse_response(https_response)

    def _check_error(self, https_response):
        response_code = https_response.status
        if response_code != 200:  # HTTP OK
            error_message = https_response.read().decode("utf-8")
            raise IOError(error_message)

    def _parse_response(self, https_response):
        json_value = json.loads(https_response.read().decode("utf-8"))
        private_to_public_addresses = dict()
        for value in json_value:
            private_address = value[self._PRIVATE_ADDRESS_PROPERTY]
            public_address = value[self._PUBLIC_ADDRESS_PROPERTY]

            public_addr = AddressHelper.address_from_str(public_address)
            # If not explicitly given, create the private address with the public addresses port
            private_addr = AddressHelper.address_from_str(private_address, public_addr.port)
            private_to_public_addresses[private_addr] = public_addr

        return private_to_public_addresses



import logging
import threading
import uuid
from collections import OrderedDict

from hazelcast import six
from hazelcast.errors import TargetDisconnectedError, IllegalStateError
from hazelcast.util import check_not_none

_logger = logging.getLogger(__name__)


class _MemberListSnapshot(object):
    __slots__ = ("version", "members")

    def __init__(self, version, members):
        self.version = version
        self.members = members


class ClientInfo(object):
    """Local information of the client.

    Attributes:
        uuid (uuid.UUID): Unique id of this client instance.
        address (hazelcast.core.Address): Local address that is used to communicate with cluster.
        name (str): Name of the client.
        labels (set[str]): Read-only set of all labels of this client.
    """

    __slots__ = ("uuid", "address", "name", "labels")

    def __init__(self, client_uuid, address, name, labels):
        self.uuid = client_uuid
        self.address = address
        self.name = name
        self.labels = labels

    def __repr__(self):
        return "ClientInfo(uuid=%s, address=%s, name=%s, labels=%s)" % (
            self.uuid,
            self.address,
            self.name,
            self.labels,
        )


_EMPTY_SNAPSHOT = _MemberListSnapshot(-1, OrderedDict())
_INITIAL_MEMBERS_TIMEOUT_SECONDS = 120


class ClusterService(object):
    """
    Cluster service for Hazelcast clients.

    It provides access to the members in the cluster
    and one can register for changes in the cluster members.
    """

    def __init__(self, internal_cluster_service):
        self._service = internal_cluster_service

    def add_listener(self, member_added=None, member_removed=None, fire_for_existing=False):
        """
        Adds a membership listener to listen for membership updates.

        It will be notified when a member is added to cluster or removed from cluster.
        There is no check for duplicate registrations, so if you register the listener
        twice, it will get events twice.

        Args:
            member_added (function): Function to be called when a member is added to the cluster.
            member_removed (function): Function to be called when a member is removed from the cluster.
            fire_for_existing (bool): Whether or not fire member_added for existing members.

        Returns:
            str: Registration id of the listener which will be used for removing this listener.
        """
        return self._service.add_listener(member_added, member_removed, fire_for_existing)

    def remove_listener(self, registration_id):
        """
        Removes the specified membership listener.

        Args:
            registration_id (str): Registration id of the listener to be removed.

        Returns:
            bool: ``True`` if the registration is removed, ``False`` otherwise.
        """
        return self._service.remove_listener(registration_id)

    def get_members(self, member_selector=None):
        """
        Lists the current members in the cluster.

        Every member in the cluster returns the members in the same order.
        To obtain the oldest member in the cluster, you can retrieve the first item in the list.

        Args:
            member_selector (function): Function to filter members to return.
                If not provided, the returned list will contain all the available cluster members.

        Returns:
            list[hazelcast.core.MemberInfo]: Current members in the cluster
        """
        return self._service.get_members(member_selector)


class _InternalClusterService(object):
    def __init__(self, client, config):
        self._client = client
        self._connection_manager = None
        self._labels = frozenset(config.labels)
        self._listeners = {}
        self._member_list_snapshot = _EMPTY_SNAPSHOT
        self._initial_list_fetched = threading.Event()

    def start(self, connection_manager, membership_listeners):
        self._connection_manager = connection_manager
        for listener in membership_listeners:
            self.add_listener(*listener)

    def get_member(self, member_uuid):
        check_not_none(uuid, "UUID must not be null")
        snapshot = self._member_list_snapshot
        return snapshot.members.get(member_uuid, None)

    def get_members(self, member_selector=None):
        snapshot = self._member_list_snapshot
        if not member_selector:
            return list(snapshot.members.values())

        members = []
        for member in six.itervalues(snapshot.members):
            if member_selector(member):
                members.append(member)
        return members

    def size(self):
        """
        Returns:
            int: Size of the cluster.
        """
        snapshot = self._member_list_snapshot
        return len(snapshot.members)

    def get_local_client(self):
        """
        Returns:
            hazelcast.cluster.ClientInfo: The client info.
        """
        connection_manager = self._connection_manager
        connection = connection_manager.get_random_connection()
        local_address = None if not connection else connection.local_address
        return ClientInfo(
            connection_manager.client_uuid, local_address, self._client.name, self._labels
        )

    def add_listener(self, member_added=None, member_removed=None, fire_for_existing=False):
        registration_id = str(uuid.uuid4())
        self._listeners[registration_id] = (member_added, member_removed)

        if fire_for_existing and member_added:
            snapshot = self._member_list_snapshot
            for member in six.itervalues(snapshot.members):
                member_added(member)

        return registration_id

    def remove_listener(self, registration_id):
        try:
            self._listeners.pop(registration_id)
            return True
        except KeyError:
            return False

    def wait_initial_member_list_fetched(self):
        """Blocks until the initial member list is fetched from the cluster.

        If it is not received within the timeout, an error is raised.

        Raises:
            IllegalStateError: If the member list could not be fetched
        """
        fetched = self._initial_list_fetched.wait(_INITIAL_MEMBERS_TIMEOUT_SECONDS)
        if not fetched:
            raise IllegalStateError("Could not get initial member list from cluster!")

    def clear_member_list_version(self):
        _logger.debug("Resetting the member list version")

        current = self._member_list_snapshot
        if current is not _EMPTY_SNAPSHOT:
            self._member_list_snapshot = _MemberListSnapshot(0, current.members)

    def clear_member_list(self):
        _logger.debug("Resetting the member list")

        current = self._member_list_snapshot
        if current is not _EMPTY_SNAPSHOT:
            previous_members = current.members
            snapshot = _MemberListSnapshot(0, {})
            self._member_list_snapshot = snapshot
            dead_members, new_members = self._detect_membership_events(
                previous_members, snapshot.members
            )
            self._fire_membership_events(dead_members, new_members)

    def handle_members_view_event(self, version, member_infos):
        snapshot = self._create_snapshot(version, member_infos)
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                "Handling new snapshot with membership version: %s, member string: %s",
                version,
                self._members_string(snapshot.members),
            )

        current = self._member_list_snapshot
        if version >= current.version:
            self._apply_new_state_and_fire_events(current, snapshot)

        if current is _EMPTY_SNAPSHOT:
            self._initial_list_fetched.set()

    def _apply_new_state_and_fire_events(self, current, snapshot):
        self._member_list_snapshot = snapshot
        dead_members, new_members = self._detect_membership_events(
            current.members, snapshot.members
        )
        self._fire_membership_events(dead_members, new_members)

    def _fire_membership_events(self, dead_members, new_members):
        # Removal events should be fired first
        for dead_member in dead_members:
            for _, handler in six.itervalues(self._listeners):
                if handler:
                    try:
                        handler(dead_member)
                    except:
                        _logger.exception("Exception in membership listener")

        for new_member in new_members:
            for handler, _ in six.itervalues(self._listeners):
                if handler:
                    try:
                        handler(new_member)
                    except:
                        _logger.exception("Exception in membership listener")

    def _detect_membership_events(self, previous_members, current_members):
        new_members = []
        dead_members = set(six.itervalues(previous_members))
        for member in six.itervalues(current_members):
            try:
                dead_members.remove(member)
            except KeyError:
                new_members.append(member)

        for dead_member in dead_members:
            connection = self._connection_manager.get_connection(dead_member.uuid)
            if connection:
                connection.close(
                    None,
                    TargetDisconnectedError(
                        "The client has closed the connection to this member, "
                        "after receiving a member left event from the cluster. "
                        "%s" % connection
                    ),
                )

        if (len(new_members) + len(dead_members)) > 0:
            if len(current_members) > 0:
                _logger.info(self._members_string(current_members))

        return dead_members, new_members

    @staticmethod
    def _members_string(members):
        n = len(members)
        return "\n\nMembers [%s] {\n\t%s\n}\n" % (n, "\n\t".join(map(str, six.itervalues(members))))

    @staticmethod
    def _create_snapshot(version, member_infos):
        new_members = OrderedDict()
        for member_info in member_infos:
            new_members[member_info.uuid] = member_info
        return _MemberListSnapshot(version, new_members)


class VectorClock(object):
    """Vector clock consisting of distinct replica logical clocks.

    The vector clock may be read from different thread but concurrent
    updates must be synchronized externally. There is no guarantee for
    concurrent updates.

    See Also:
        https://en.wikipedia.org/wiki/Vector_clock
    """

    def __init__(self):
        self._replica_timestamps = {}

    def is_after(self, other):
        """Returns ``True`` if this vector clock is causally strictly after the
        provided vector clock. This means that it the provided clock is neither
        equal to, greater than or concurrent to this vector clock.

        Args:
            other (hazelcast.cluster.VectorClock): Vector clock to be compared

        Returns:
            bool: ``True`` if this vector clock is strictly after the other vector clock, ``False`` otherwise.
        """
        any_timestamp_greater = False
        for replica_id, other_timestamp in other.entry_set():
            local_timestamp = self._replica_timestamps.get(replica_id)

            if local_timestamp is None or local_timestamp < other_timestamp:
                return False
            elif local_timestamp > other_timestamp:
                any_timestamp_greater = True

        # there is at least one local timestamp greater or local vector clock has additional timestamps
        return any_timestamp_greater or other.size() < self.size()

    def set_replica_timestamp(self, replica_id, timestamp):
        """Sets the logical timestamp for the given replica ID.

        Args:
            replica_id (str): Replica ID.
            timestamp (int): Timestamp for the given replica ID.
        """
        self._replica_timestamps[replica_id] = timestamp

    def entry_set(self):
        """Returns the entry set of the replica timestamps in a format of list of tuples.

        Each tuple contains the replica ID and the timestamp associated with it.

        Returns:
            list: List of tuples.
        """
        return list(self._replica_timestamps.items())

    def size(self):
        """Returns the number of timestamps that are in the replica timestamps dictionary.

        Returns:
            int: Number of timestamps in the replica timestamps.
        """
        return len(self._replica_timestamps)
